# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "14/02/2020"


import logging
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.gui.reconstruction.nabu import check
from contextlib import AbstractContextManager
from silx.gui import qt
from tomwer.synctools.stacks.reconstruction.nabu import NabuVolumeProcessStack
from tomwer.core.process.reconstruction.nabu.nabuvolume import NabuVolume
from tomwer.gui.reconstruction.nabu.volume import NabuVolumeWindow
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.scan.futurescan import FutureTomwerScan
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from orangewidget.settings import Setting
from orangewidget import gui
from orangewidget.widget import Input, Output
from ...orange.managedprocess import SuperviseOW
from tomwer.core.process.reconstruction.nabu import utils as nabu_utils
from tomwer.core.utils.char import BETA_CHAR, DELTA_CHAR
from collections.abc import Iterable
from tomwer.utils import docstring
from tomwer.core import settings
import tomwer.core.process.reconstruction.nabu.nabuvolume
import functools
import copy

_logger = logging.getLogger(__name__)


class NabuVolumeOW(WidgetLongProcessing, SuperviseOW):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "nabu volume reconstruction"
    id = "orange.widgets.tomwer.reconstruction.NabuVolumeOW.NabuVolumeOW"
    description = (
        "This widget will call nabu for running a reconstruction " "on a volume"
    )
    icon = "icons/nabu_3d.svg"
    priority = 15
    keywords = ["tomography", "nabu", "reconstruction", "volume"]

    ewokstaskclass = tomwer.core.process.reconstruction.nabu.nabuvolume.NabuVolume

    want_main_area = True
    resizing_enabled = True
    compress_signal = False
    allows_cycle = True

    _rpSetting = Setting(dict())
    # kept for compatibility

    static_input = Setting(
        {"data": None, "nabu_volume_params": None, "nabu_params": None}
    )

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    TIMEOUT = 30

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=False,
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

        future_out = Output(
            name="future_data",
            type=FutureTomwerScan,
            doc="data with some remote processing",
        )

    class DialogCM(AbstractContextManager):
        """Simple context manager to hida / show button dialogs"""

        def __init__(self, dialogButtonsBox):
            self._dialogButtonsBox = dialogButtonsBox

        def __enter__(self):
            self._dialogButtonsBox.show()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._dialogButtonsBox.hide()

    def __init__(self, parent=None):
        """
        Widget which read the .hdf5 generated by octave and modify it.
        Then run a subprocess to call octave and run ftseries

        :param bool _connect_handler: True if we want to store the modifications
                                      on the setting. Need for unit test since
                                      keep alive qt widgets.
        :param recons_params: reconsparameter to be used by the FTWidget.
                              If None, some will be created.
        :type: :class:`QReconsParams`
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._slurmCluster = None
        self.__exec_for_ci = False
        # processing tool
        self._processingStack = NabuVolumeProcessStack(self, process_id=self.process_id)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        # main widget
        self._nabuWidget = NabuVolumeWindow(parent=self)
        _layout.addWidget(self._nabuWidget)
        # add button to validate when change reconstruction parameters is called
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        _layout.addWidget(self._buttons)

        # set up
        self._buttons.hide()

        # load settings
        nabu_volume_params = self.static_input.get("nabu_volume_params", None)
        if nabu_volume_params is None:
            nabu_volume_params = self._rpSetting
        if nabu_volume_params != dict():
            try:
                self._nabuWidget.setConfiguration(nabu_volume_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        # connect signal / slot
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._nabuWidget.sigConfigChanged.connect(self._updateSettingsVals)

    @Inputs.data
    def process(self, scan: TomwerScanBase):
        assert isinstance(scan, (TomwerScanBase, type(None)))
        if scan is None:
            return
        scan_ = copy.copy(scan)
        scan_.clear_latest_vol_reconstructions()

        if isinstance(scan, HDF5TomoScan):
            if not check.check_flat_series(
                scan_, logger=_logger, user_input=settings.isOnLbsram(scan=scan_)
            ):
                self.Outputs.data.send(scan_)
                return
            if not check.check_dark_series(
                scan_, logger=_logger, user_input=settings.isOnLbsram(scan=scan_)
            ):
                self.Outputs.data.send(scan_)
                return

        # insure we are able to reconstruct
        if scan.nabu_recons_params in ({}, None):
            _logger.error(
                "No reconstruction parameters found from nabu "
                "slices for {}. You should first run slice "
                "reconstruction prior to volume reconstruction"
                "".format(str(scan))
            )
            self.Outputs.data.send(scan)
            self.sigScanReady.emit(scan)
            return

        config = scan.nabu_recons_params
        if "phase" in config and "delta_beta" in config["phase"]:
            pag_dbs = config["phase"]["delta_beta"]
            if isinstance(pag_dbs, str):
                try:
                    pag_dbs = nabu_utils.retrieve_lst_of_value_from_str(
                        config["phase"]["delta_beta"], type_=float
                    )
                except Exception:
                    pass
            if len(pag_dbs) > 1:
                _logger.warning(
                    "Several value found for {} / {}. Volume"
                    "reconstruction take one at most."
                    "".format(DELTA_CHAR, BETA_CHAR)
                )
                timeout = NabuVolumeOW.TIMEOUT if settings.isOnLbsram(scan) else None
                self._dialogDB = _DeltaBetaSelectorDialog(
                    values=pag_dbs, parent=None, timeout=timeout
                )
                self._dialogDB.setModal(False)
                self._callbackDB = functools.partial(
                    self._updateDB, scan, self._dialogDB
                )
                self._dialogDB.accepted.connect(self._callbackDB)
                self._callbackTimeout = functools.partial(self._skipProcessing, scan)
                self._dialogDB.timeoutReached.connect(self._callbackTimeout)
                self._dialogDB.show()
                return

        self._processingStack.add(scan_, self.getConfiguration())

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        if (
            dataset.axis_params is None
            or dataset.axis_params.relative_cor_value is None
        ):
            # try to retrieve last computed cor value from nabu process
            r_cor = NabuVolume.retrieve_last_relative_cor(dataset)
            if dataset.axis_params is None:
                from tomwer.synctools.axis import QAxisRP

                dataset.axis_params = QAxisRP()
            try:
                dataset.axis_params.set_relative_value(
                    float(r_cor) - dataset.dim_1 / 2.0
                )
            except Exception:
                pass

        self.process(dataset)

    @Inputs.cluster_in
    def setCluster(self, cluster):
        self._slurmCluster = cluster

    def _updateDB(self, scan, dialog):
        db = dialog.getSelectedValue()
        if db is not None:
            try:
                scan.nabu_recons_params["phase"]["delta_beta"] = db
            except Exception as e:
                logging.error(e)
            else:
                self.process(scan=scan)

    def _endProcessing(self, scan, future_scan):
        WidgetLongProcessing._endProcessing(self, scan)
        if scan is not None:
            self.Outputs.data.send(scan)
            self.sigScanReady.emit(scan)
        if future_scan is not None:
            self.Outputs.future_out.send(future_scan)

    def setDryRun(self, dry_run):
        self._processingStack.setDryRun(dry_run)

    def _ciExec(self):
        self.activateWindow()
        self.raise_()
        self.show()

    def _replaceExec_(self):
        """used for CI, replace the exec_ call ny"""
        self.__exec_for_ci = True

    def _updateSettingsVals(self):
        self._rpSetting = self.getConfiguration()
        # kept for compatibility
        self.static_input = {
            "data": None,
            "nabu_volume_params": self.getConfiguration(),
            "nabu_params": None,
        }

    def _skipProcessing(self, scan):
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(scan)

    def getConfiguration(self):
        config = self._nabuWidget.getConfiguration()
        config["cluster_config"] = self._slurmCluster
        return config

    def setConfiguration(self, config):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        self._nabuWidget.setConfiguration(config=config)


class _DeltaBetaSelectorDialog(qt.QDialog):
    timeoutReached = qt.Signal()

    def __init__(self, values, parent=None, timeout=None):
        """

        :param values:
        :param parent:
        :param timeout: if a timeout is provided once reach this will
                        automatically reject the delta / beta selection.
                        This is needed when on lbsral to avoid 'locking'
                        a reconstruction. In sec.
        """
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self.mainWidget = _DeltaBetaSelector(parent=self, values=values)
        self.layout().addWidget(self.mainWidget)

        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)
        self._timeout = timeout
        if self._timeout is None:
            self.setWindowTitle(
                "Select one value for {} / {}".format(DELTA_CHAR, BETA_CHAR)
            )
        else:
            self.setWindowTitle(
                "Select one value for {} / {}. (close automatically in {} "
                "sec.)".format(DELTA_CHAR, BETA_CHAR, self._timeout)
            )

        # connect signal / slot
        self._buttons.accepted.connect(self.accept)
        self._buttons.rejected.connect(self.reject)

        # expose API
        self.getSelectedValue = self.mainWidget.getSelectedValue

        # add timers
        if timeout is not None:
            self._timer = qt.QTimer()
            self._timer.timeout.connect(self.reject)
            self._timer.start(timeout * 1000)
            self._displayTimer = qt.QTimer()
            self._displayTimer.timeout.connect(self._updateTitle)
            self._displayTimer.start(1000)
        else:
            self._timer = None

    def reject(self):
        if self._timer:
            self._timer.stop()
            self._timer.timeout.disconnect(self.reject)
            self._displayTimer.stop()
            self._displayTimer.timeout.disconnect(self._updateTitle)
            self._timer = None

        qt.QDialog.reject(self)

    def _updateTitle(self):
        self._timeout = self._timeout - 1
        if self._timeout <= 0:
            self.timeoutReached.emit()
            self.reject()
        else:
            self.setWindowTitle(
                "Select one value for {} / {}. (close automatically in {} "
                "sec.)".format(DELTA_CHAR, BETA_CHAR, self._timeout)
            )
            self._displayTimer.start(1000)


class _DeltaBetaSelector(qt.QTableWidget):
    """Widget used to select a value of delta beta if several provided"""

    def __init__(self, values, parent=None):
        qt.QTableWidget.__init__(self, parent)
        self.setSelectionMode(qt.QAbstractItemView.SingleSelection)
        self.setHorizontalHeaderLabels([DELTA_CHAR + " / " + BETA_CHAR])
        self.setRowCount(0)
        self.setColumnCount(1)
        self.verticalHeader().hide()
        if hasattr(self.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            self.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            self.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)
        self.setAcceptDrops(False)

        # set up
        self.setValues(values=values)

    def setValues(self, values: Iterable):
        self.setHorizontalHeaderLabels([DELTA_CHAR + " / " + BETA_CHAR])
        self.setRowCount(len(values))
        self.setColumnCount(1)
        for i_value, value in enumerate(values):
            _item = qt.QTableWidgetItem()
            _item.setText(str(value))
            _item.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
            self.setItem(i_value, 0, _item)
            _item.setSelected(i_value == 0)

    def getSelectedValue(self):
        sel = None
        for item in self.selectedItems():
            sel = item.text()
        return sel



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "21/03/2019"


import unittest
import tempfile
import platform
from tomwer.core.process.reconstruction.ftseries.params.paganin import (
    PaganinRP,
    PaganinMode,
)
from tomwer.core.process.reconstruction.darkref.params import (
    DKRFRP,
    Method as Dkrf_Method,
)
from tomwer.core.process.reconstruction.ftseries.params.ft import FTRP
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from tomwer.core.utils import ftseriesutils
import os
import pytest


class TestReconsParamSet(unittest.TestCase):
    """
    test the generation of unique reconstruction parameters list
    """

    def testReconsParamListPaganin(self):
        """Test that several values of paganin db, db2 are bringing several
        reconstructions parameters set"""
        paganin_param = PaganinRP()
        paganin_param.db = [5.0, 6.0]
        paganin_param.db2 = [1.0, 1.0, 2.0]
        recons_list = paganin_param.to_unique_recons_set(as_to_dict=False)
        self.assertEqual(len(recons_list), 4)

    def testReconsParamListDKRF(self):
        dkrf = DKRFRP()
        for as_to_dict, th_value in zip((True, False), ("Median", Dkrf_Method.median)):
            with self.subTest(as_to_dict=as_to_dict):
                _unique = dkrf.to_unique_recons_set(as_to_dict=as_to_dict)
                self.assertEqual(len(_unique), 1)
                self.assertEqual(_unique[0]["REFSCAL"], th_value)

    def testPagReconsParamList(self):
        """Same test as testReconsParamListPaganin but from an instance of
        ReconsParams"""
        paganin_param = PaganinRP()
        paganin_param.db = [5.0, 6.0]
        paganin_param.db2 = [1.0, 1.0, 2.0]

        for empty in (True, False):
            with self.subTest(empty=empty):
                rp = ReconsParams(empty=empty)
                rp.paganin = paganin_param
                recons_list = rp.to_unique_recons_set(as_to_dict=True)
                self.assertEqual(len(recons_list), 4)
                for param_set in recons_list:
                    self.assertTrue(isinstance(param_set, dict))

    @pytest.mark.skipif(
        platform.system() == "Windows", reason="No rights on conda installation"
    )
    def testReconsParamList(self):
        """Same test as testReconsParamListPaganin but from an instance of
        ReconsParams"""
        paganin_param = PaganinRP()
        paganin_param.db = [5.0, 6.0]
        paganin_param.db2 = [1.0, 1.0, 2.0]

        fp_param = FTRP()
        fp_param.correct_spikes_threshold = (True, False)

        recons_params = ReconsParams()
        recons_params.ft = fp_param
        recons_params.paganin = paganin_param

        recons_list = recons_params.to_unique_recons_set(as_to_dict=True)
        self.assertEqual(len(recons_list), 8)
        for param_set in recons_list:
            self.assertTrue(isinstance(param_set, dict))
            tmp_file = tempfile.NamedTemporaryFile(suffix=".h5")
            with tmp_file as tmpH5File:
                # test writing h5 file
                ftseriesutils.saveH5File(
                    h5File=tmpH5File.name, structs=param_set, displayInfo=False
                )
            if os.path.exists(tmp_file.name):
                os.remove(tmp_file.name)

    def testSingleValuesReconsParams(self):
        """Make sure that a default instance of ReconsParams with only
        single values generate a single set of reconstruction parameters"""
        recons_params = ReconsParams()
        rp_list = recons_params.to_unique_recons_set(as_to_dict=True)
        self.assertEqual(len(rp_list), 1)
        self.assertTrue(isinstance(rp_list[0], dict))

    def testParamsSetWithSlicesAndPaganin(self):
        """Test that the `_UNSPLIT_KEYS` is correctly taking into account"""
        ft_param = FTRP()
        ft_param.fixed_slice = (12, 16)

        self.assertEqual(len(ft_param.to_unique_recons_set()), 1)


class TestReconsParamCopy(unittest.TestCase):
    """Test copy of the reconstruction parameters"""

    def setUp(self):
        self._rp_empty = ReconsParams(empty=True)
        self._rp_full = ReconsParams(empty=False)
        assert self._rp_full.paganin.mode is PaganinMode.off
        self._pag_rp = PaganinRP()
        self._pag_rp.mode = PaganinMode.on
        self._dkrf_rp = DKRFRP()
        assert self._rp_full.dkrf.dark_calc_method is Dkrf_Method.average
        self._dkrf_rp.dark_calc_method = Dkrf_Method.none

    def testCopy(self):
        """make sure copy function is valid"""
        for rp in (self._rp_empty, self._rp_full):
            rp.copy(self._pag_rp)
            rp.copy(self._dkrf_rp)
            if rp is self._rp_empty:
                self.assertTrue(rp.paganin is not None)
                self.assertTrue(rp.beam_geo is None)
                self.assertTrue(rp.pyhst is None)
            self.assertTrue(rp.paganin is not self._pag_rp)
            self.assertTrue(rp.dkrf is not self._dkrf_rp)
            self.assertTrue(rp.paganin.mode is PaganinMode.on)
            self.assertTrue(rp.dkrf.dark_calc_method is Dkrf_Method.none)
            self.assertEqual(len(rp.to_unique_recons_set()), 1)

            self._pag_rp.copy(rp.paganin)
            self.assertEqual(len(rp.paganin.to_unique_recons_set()), 1)
            self._dkrf_rp.copy(rp.dkrf)
            self.assertEqual(len(rp.dkrf.to_unique_recons_set()), 1)

    def testReconsParamsUnmanagedParamsCopy(self):
        self._rp_full._add_unmanaged_param(param="TEST", value=12)
        self._rp_empty.copy(self._rp_full)
        self.assertTrue("TEST" in self._rp_empty.unmanaged_params)
        self.assertTrue("TEST" in self._rp_empty.all_params)
        self.assertEqual(self._rp_empty["TEST"], 12)

    def testFTRPUnmanagedParamsCopy(self):
        rp_1 = FTRP()
        rp_2 = FTRP()
        rp_2._add_unmanaged_param("TOTO", 12)
        rp_2._add_unmanaged_param("TATA", "dadsad")
        rp_2._add_unmanaged_param("KAWA", (23, 5))
        rp_1.copy(rp_2)
        self.assertTrue(rp_1.to_dict() == rp_2.to_dict())



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "30/07/2020"


import unittest
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
from tomwer.gui.control.datavalidator import DataValidator


class TestValidatorGUI(TestCaseQt):
    """Test that the ImageStackWidget can be load and remove without any issue"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = DataValidator(None)

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        unittest.TestCase.tearDown(self)

    def test(self):
        """Make sur the addLeaf and clear functions are working"""
        pass



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to axis calculation using radios
"""

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/02/2019"

from silx.gui import qt
import logging
from tomwer.core.utils import image
import numpy
import enum
import os
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.synctools.axis import QAxisRP
from tomwer.core.process.reconstruction.axis import mode as axis_mode
from tomwer.core.process.reconstruction.axis.anglemode import CorAngleMode
from tomwer.core.process.reconstruction.axis.params import (
    AxisCalculationInput,
    DEFAULT_CMP_THETA,
    DEFAULT_CMP_SUBSAMPLING_Y,
    DEFAULT_CMP_OVERSAMPLING,
    DEFAULT_CMP_TAKE_LOG,
)
from silx.io.url import DataUrl
import scipy.signal
from .CompareImages import CompareImages
from tomwer.gui.utils.buttons import PadlockButton
from silx.utils.deprecation import deprecated
from typing import Union
from typing import Optional

_logger = logging.getLogger(__name__)


class RadioAxisWindow(qt.QMainWindow):
    """
    QMainWindow for defining the rotation axis

    .. snapshotqt:: img/AxisWindow.png

        from tomwer.gui.ftserie.axis import AxisWindow
        from tomwer.synctools.ftseries import QReconsParams
        import scipy.misc
        import scipy.ndimage

        recons_params = QReconsParams()
        widget = AxisWindow(recons_params)
        imgA = scipy.misc.ascent()
        imgB = scipy.ndimage.affine_transform(imgA, (1, 1), offset=(0, 10))

        widget.setImages(imgA=imgA, imgB=imgB, flipB=False)
        widget.show()

    :raises ValueError: given axis is not an instance of _QAxisRP
    """

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the lock on the mode change"""

    sigPositionChanged = qt.Signal(tuple)
    """signal emitted when the center of rotation center change"""

    def __init__(self, axis, parent=None):
        qt.QMainWindow.__init__(self, parent)
        if isinstance(axis, QAxisRP):
            self.__recons_params = axis
        else:
            raise TypeError("axis should be an instance of _QAxisRP")

        self._imgA = None
        self._imgB = None
        self._shiftedImgA = None
        self._flipB = True
        """Option if we want to flip the image B"""
        self.__callback = None
        self._scan = None
        self._axis_params = None
        self._lastManualFlip = None
        """Cache for the last user entry for manual flip"""
        self._lastXShift = None
        # cache to know if the x shift has changed since
        self._lastYShift = None
        # cache to know if the y shift has changed
        self._lastXOrigin = None
        # cache to know if the x origin has changed since
        self._lastYOrigin = None
        # cache to know if the y origin has changed since

        self.setWindowFlags(qt.Qt.Widget)
        self._plot = CompareImages(parent=self)
        self._plot.setAutoResetZoom(False)
        _mode = CompareImages.VisualizationMode.COMPOSITE_A_MINUS_B
        self._plot.setVisualizationMode(_mode)
        self.setCentralWidget(self._plot)

        self._dockWidgetCtrl = qt.QDockWidget(parent=self)
        self._dockWidgetCtrl.layout().setContentsMargins(0, 0, 0, 0)
        self._dockWidgetCtrl.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._controlWidget = _AxisManual(
            parent=self, reconsParams=self.__recons_params
        )
        self._dockWidgetCtrl.setWidget(self._controlWidget)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockWidgetCtrl)

        # expose API
        self.getXShift = self._controlWidget.getXShift
        self.getYShift = self._controlWidget.getYShift
        self._setShift = self._controlWidget.setShift
        self.setXShift = self._controlWidget.setXShift
        self.setYShift = self._controlWidget.setYShift
        self.getShiftStep = self._controlWidget.getShiftStep
        self.setShiftStep = self._controlWidget.setShiftStep
        self.getAxis = self._controlWidget.getAxis
        self.getMode = self._controlWidget.getMode
        self.setModeLock = self._controlWidget.setModeLock
        self.getEstimatedCor = self._controlWidget.getEstimatedCor

        # signal / slot connection
        self._controlWidget.sigShiftChanged.connect(self._updateShift)
        self._controlWidget.sigShiftChanged.connect(self._corChanged)
        self._controlWidget.sigRoiChanged.connect(self._updateShift)
        self._controlWidget.sigAuto.connect(self._updateAuto)
        self._controlWidget.sigModeChanged.connect(self.setMode)
        self._controlWidget.sigModeLockChanged.connect(self._modeLockChanged)
        self._controlWidget.sigResetZoomRequested.connect(self._resetZoomPlot)
        self._controlWidget.sigSubsamplingChanged.connect(self._updateSubSampling)
        self._controlWidget.sigUrlChanged.connect(self._urlChanged)

        # adapt gui to the axis value
        self.setReconsParams(axis=self.__recons_params)

    def manual_uses_full_image(self, value):
        self._controlWidget.manual_uses_full_image(value)

    def _modeLockChanged(self, lock):
        self.sigLockModeChanged.emit(lock)

    def _corChanged(self):
        self.sigPositionChanged.emit((self.getXShift(), self.getYShift()))

    def getPlot(self):
        return self._plot

    def _resetZoomPlot(self):
        self._plot.getPlot().resetZoom()

    def getSettingsWidget(self):
        return self._controlWidget

    def setMode(self, mode):
        """
        Define the mode to use for radio axis

        :param mode:
        :return:
        """
        mode = axis_mode.AxisMode.from_value(mode)
        old_control = self._controlWidget.blockSignals(True)
        old_axis = self._axis_params.blockSignals(True)
        self._controlWidget.setMode(mode)
        self._controlWidget.blockSignals(old_control)
        self._axis_params.blockSignals(old_axis)

    def setEstimatedCor(self, value):
        self._controlWidget.setEstimatedCor(value=value)

    def getEstimatedCor(self):
        return self._controlWidget.getEstimatedCor()

    def _setModeLockFrmSettings(self, lock):
        # only lock the push button
        old = self.blockSignals(True)
        self._controlWidget._mainWidget._calculationWidget._lockMethodPB.setLock(lock)
        self.blockSignals(old)

    def getROIDims(self):
        if self.getMode() == axis_mode.AxisMode.manual:
            return self._controlWidget.getROIDims()
        else:
            return None

    def getROIOrigin(self):
        if self.getMode() == axis_mode.AxisMode.manual:
            return self._controlWidget.getROIOrigin()
        else:
            return None

    def getImgSubsampling(self):
        return self._controlWidget.getImgSubsampling()

    def _computationRequested(self):
        self.sigComputationRequested.emit()

    def setLocked(self, locked):
        old = self.blockSignals(True)
        if self._axis_params.mode not in (
            axis_mode.AxisMode.read,
            axis_mode.AxisMode.manual,
        ):
            self._axis_params.mode = axis_mode.AxisMode.manual
        self._controlWidget.setLocked(locked)
        self.blockSignals(old)
        self.sigAxisEditionLocked.emit(locked)

    def isModeLock(self):
        return self._controlWidget.isModeLock()

    def _validated(self):
        """callback when the validate button is activated"""
        self.sigApply.emit()

    def _setRadio2Flip(self, checked):
        self._plot.setRadio2Flip(checked)

    def _flipChanged(self, checked):
        if self.getMode() == axis_mode.AxisMode.manual:
            self._lastManualFlip = self._plot.isRadio2Flip()

        if checked == self._flipB:
            return
        else:
            self._flipB = checked
            self._updatePlot()

    def setReconsParams(self, axis):
        """

        :param AxisRP axis: axis to edit
        :return:
        """
        assert isinstance(axis, QAxisRP)
        self._axis_params = axis
        old = self.blockSignals(True)
        self.resetShift()
        self._controlWidget.setAxis(axis)
        self.blockSignals(old)

    def setScan(self, scan):
        """
        Update the interface concerning the given scan. Try to display the
        radios for angle 0 and 180.

        :param scan: scan for which we want the axis updated.
        :type scan: Union[str, tomwer.core.scan.TomoBase]
        """
        self.clear()
        _scan = scan
        if type(scan) is str:
            try:
                _scan = ScanFactory.create_scan_object(scan)
            except ValueError:
                raise ValueError("Fail to discover a valid scan in %s" % scan)
        elif not isinstance(_scan, TomwerScanBase):
            raise ValueError(
                "type of %s (%s) is invalid, scan should be a "
                "file/dir path or an instance of ScanBase" % (scan, type(scan))
            )
        assert isinstance(_scan, TomwerScanBase)

        if _scan.axis_params is None:
            _scan.axis_params = QAxisRP()

        if self._scan is not None:
            self._scan.axis_params.sigAxisUrlChanged.disconnect(self._updatePlot)

        # update visualization
        self._scan = _scan
        self._scan.axis_params.sigAxisUrlChanged.connect(self._updatePlot)
        self._controlWidget.setScan(scan=self._scan)
        self._updatePlot()
        self.getPlot().getPlot().resetZoom()

    def _updatePlot(self):
        if self._scan is None:
            return
        self._urlChanged()

    def _urlChanged(self):
        old = self.blockSignals(True)
        coreAngleMode = CorAngleMode.from_value(self.__recons_params.angle_mode)
        if self._scan is None:
            return
        axis_rp = self._scan.axis_params
        if coreAngleMode is CorAngleMode.manual_selection:
            manual_sel_widget = (
                self._controlWidget._mainWidget._inputWidget._angleModeWidget._manualFrameSelection
            )
            urls = manual_sel_widget.getFramesUrl(as_txt=False)
            axis_rp.axis_url_1, axis_rp.axis_url_2 = urls
            axis_rp.flip_lr = manual_sel_widget.isFrame2LRFLip()
        else:
            axis_rp.flip_lr = True
            res = self._scan.getRadiosForAxisCalc(mode=coreAngleMode)
            axis_rp.axis_url_1 = res[0]
            axis_rp.axis_url_2 = res[1]

        if axis_rp.n_url() < 2:
            _logger.error("Fail to detect radio for axis calculation")
        elif axis_rp.axis_url_1.url:
            # if necessary normalize data
            axis_rp.axis_url_1.normalize_data(self._scan, log_=False)
            axis_rp.axis_url_2.normalize_data(self._scan, log_=False)

            paganin = self.__recons_params.paganin_preproc
            # check if normed
            if paganin:
                imgA = axis_rp.axis_url_1.normalized_data_paganin
                imgB = axis_rp.axis_url_2.normalized_data_paganin
            else:
                imgA = axis_rp.axis_url_1.normalized_data
                imgB = axis_rp.axis_url_2.normalized_data
            assert imgA is not None
            assert imgB is not None
            self.setImages(imgA=imgA, imgB=imgB, flipB=axis_rp.flip_lr)
        else:
            _logger.error(
                "fail to find radios for angle 0 and 180. Unable to update axis gui"
            )
        self.blockSignals(old)

    def clear(self):
        if self._scan is not None:
            self._scan.axis_params.sigAxisUrlChanged.disconnect(self._updatePlot)
        self._scan = None

    def setImages(self, imgA, imgB, flipB):
        """

        :warning: does not reset the shift when change images

        :param numpy.array imgA: first image to compare. Will be the one shifted
        :param numpy.array imgB: second image to compare
        :param bool flipB: True if the image B has to be flipped
        :param bool paganin: True to apply paganin phase retrieval
        """
        assert imgA is not None
        assert imgB is not None
        _imgA = imgA
        _imgB = imgB

        if _imgA.shape != _imgB.shape:
            _logger.error(
                "The two provided images have incoherent shapes "
                "(%s vs %s)" % (_imgA.shape, _imgB.shape)
            )
        elif _imgA.ndim != 2:
            _logger.error("Image shape are not 2 dimensional")
        else:
            self._imgA = _imgA
            self._imgB = _imgB
            self._flipB = flipB

            self._controlWidget._roiControl.setLimits(
                width=self._imgA.shape[1], height=self._imgA.shape[0]
            )
            self._updateShift()

    def _updateSubSampling(self):
        self._updateShift()
        self.getPlot().getPlot().resetZoom()

    def _updateShift(self, xShift=None, yShift=None):
        if self._imgA is None or self._imgB is None:
            return
        xShift = xShift or self.getXShift()
        yShift = yShift or self.getYShift()

        # TODO: we might avoid flipping image at each new x_shift...
        _imgA, _imgB = self._getRawImages()
        # apply shift
        if xShift == 0.0 and yShift == 0.0:
            self._shiftedImgA = _imgA
            self._shiftedImgB = _imgB
        else:
            try:
                cval_imgA = _imgA.min()
                cval_imgB = _imgB.min()
            except ValueError:
                _logger.warning("enable to retrieve imgA.min() and / or" "imgB.min().")
                cval_imgA = 0
                cval_imgB = 0
            try:
                self._shiftedImgA = image.shift_img(
                    data=_imgA,
                    dx=-self.getXShift() / self.getImgSubsampling(),
                    dy=self.getYShift() / self.getImgSubsampling(),
                    cval=cval_imgA,
                )
                self._shiftedImgB = image.shift_img(
                    data=_imgB,
                    dx=self.getXShift() / self.getImgSubsampling(),
                    dy=self.getYShift() / self.getImgSubsampling(),
                    cval=cval_imgB,
                )
            except ValueError as e:
                _logger.error(e)
                self._shiftedImgA = _imgA
                self._shiftedImgB = _imgB
        old = self.blockSignals(True)
        try:
            try:
                self._plot.setData(
                    image1=self._shiftedImgA,
                    image2=self._shiftedImgB,
                    updateColormap=False,
                )
            except TypeError:
                self._plot.setData(image1=self._shiftedImgA, image2=self._shiftedImgB)
        except ValueError:
            _logger.warning(
                "Unable to set images. Maybe there is some "
                "incomplete dataset or an issue with "
                "normalization."
            )
        roi_origin = self.getROIOrigin()
        if roi_origin is not None:
            x_origin, y_origin = roi_origin
        else:
            x_origin = y_origin = None
        self._lastXShift = xShift
        self._lastYShift = yShift
        self._lastXOrigin = x_origin
        self._lastYOrigin = y_origin
        self.blockSignals(old)

    def _getRawImages(self):
        def selectROI(data, width, height, x_origin, y_origin, subsampling):
            assert subsampling > 0
            x_min = x_origin - width // 2
            x_max = x_origin + width // 2
            y_min = y_origin - height // 2
            y_max = y_origin + height // 2
            return data[y_min:y_max:subsampling, x_min:x_max:subsampling]

        # get images and apply ROI if any
        _roi_dims = self.getROIDims()
        _origin = self.getROIOrigin()
        subsampling = self.getImgSubsampling()
        _imgA = self._imgA
        _imgB = self._imgB
        # flip image B
        _imgB = numpy.fliplr(_imgB) if self._flipB else _imgB
        if _roi_dims is not None:
            assert type(_roi_dims) is tuple, f"invalide roi value {_roi_dims}"
            _imgA = selectROI(
                _imgA,
                width=_roi_dims[0],
                height=_roi_dims[1],
                x_origin=_origin[0],
                y_origin=_origin[1],
                subsampling=subsampling,
            )
            _imgB = selectROI(
                _imgB,
                width=_roi_dims[0],
                height=_roi_dims[1],
                x_origin=_origin[0],
                y_origin=_origin[1],
                subsampling=subsampling,
            )
        return _imgA, _imgB

    def _updateAuto(self):
        _imgA, _imgB = self._getRawImages()
        correlation = scipy.signal.correlate2d(in1=_imgA, in2=_imgB)
        y, x = numpy.unravel_index(numpy.argmax(correlation), correlation.shape)
        self._setShift(x=x, y=y)

    def resetShift(self):
        self._controlWidget.blockSignals(True)
        self._controlWidget.reset()
        self._controlWidget.blockSignals(False)
        if self._imgA is not None and self._imgB is not None:
            self.setImages(imgA=self._imgA, imgB=self._imgB, flipB=self._flipB)


class _AxisRead(qt.QWidget):
    """Widget to select a position value from a file"""

    sigFileChanged = qt.Signal(str)

    def __init__(self, parent, axis=None):
        qt.QWidget.__init__(self, parent)
        self._axis = None
        if axis:
            self.setAxis(axis)
        self.setLayout(qt.QHBoxLayout())

        self.layout().addWidget(qt.QLabel("File", parent=self))
        self._filePathQLE = qt.QLineEdit("", parent=self)
        self.layout().addWidget(self._filePathQLE)
        self._fileSelPB = qt.QPushButton("select", parent=self)
        self.layout().addWidget(self._fileSelPB)

        # connect signal / slot
        self._fileSelPB.pressed.connect(self._selectFile)
        self._filePathQLE.textChanged.connect(self._fileChanged)

    def setAxis(self, axis):
        assert isinstance(axis, QAxisRP)
        self._axis = axis

    def _selectFile(self):
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.ExistingFile)

        if not dialog.exec_():
            dialog.close()
            return

        _file_path = dialog.selectedFiles()[0]
        _logger.info("user select file %s for reading position value" % _file_path)
        self._filePathQLE.setText(dialog.selectedFiles()[0])

    def _fileChanged(self, file_path):
        """callback when the line edit (containing the file path) changed"""
        if self._axis and os.path.isfile(file_path):
            self._axis.set_position_frm_par_file(file_path, force=True)


class _AxisManual(qt.QWidget):
    """
    Widget to define the shift to apply on an image
    """

    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted when requested shift changed. Parameter is x, y"""

    sigModeLockChanged = qt.Signal(bool)
    """Signal emitted when the mode is lock or unlock"""

    sigResetZoomRequested = qt.Signal()
    """Signal emitted when request a zoom reset from the plot"""

    sigSubsamplingChanged = qt.Signal()
    """Signal emitted when subsampling change"""

    sigUrlChanged = qt.Signal()
    """Signal emit when frames urls changed"""

    def __init__(self, parent, reconsParams):
        assert isinstance(reconsParams, QAxisRP)
        qt.QWidget.__init__(self, parent)
        self._xShift = 0
        self._yShift = 0
        self._recons_params = reconsParams or QAxisRP()
        self._axis = None

        self.setLayout(qt.QVBoxLayout())

        self._manualSelectionWidget = _AxisManualSelection(
            parent=self, shift_mode=ShiftMode.x_only
        )
        self._manualSelectionWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._readFileSelWidget = _AxisRead(parent=self)
        self._readFileSelWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._displacementSelector = self._manualSelectionWidget._displacementSelector
        self._shiftControl = self._manualSelectionWidget._shiftControl
        self._roiControl = self._manualSelectionWidget._roiControl
        self._imgOpts = self._manualSelectionWidget._imgOpts

        self._mainWidget = AxisTabWidget(
            parent=self,
            mode_dependant_widget=self._manualSelectionWidget,
            read_file_sel_widget=self._readFileSelWidget,
            recons_params=self._recons_params,
        )

        self.layout().addWidget(self._mainWidget)

        # signal / slot connection
        self._shiftControl.sigShiftLeft.connect(self._incrementLeftShift)
        self._shiftControl.sigShiftRight.connect(self._incrementRightShift)
        self._shiftControl.sigShiftTop.connect(self._incrementTopShift)
        self._shiftControl.sigShiftBottom.connect(self._incrementBottomShift)
        self._shiftControl.sigReset.connect(self._resetShift)
        self._shiftControl.sigShiftChanged.connect(self._setShiftAndSignal)
        self._mainWidget.sigLockModeChanged.connect(self._modeLockChanged)
        self._manualSelectionWidget.sigResetZoomRequested.connect(
            self._requestZoomReset
        )
        self._imgOpts.sigSubsamplingChanged.connect(self._subsamplingChanged)
        self._mainWidget.sigUrlChanged.connect(self._urlChanged)

        # expose API
        self.getShiftStep = self._displacementSelector.getShiftStep
        self.setShiftStep = self._displacementSelector.setShiftStep
        self.sigRoiChanged = self._roiControl.sigRoiChanged
        self.sigAuto = self._shiftControl.sigAuto
        self.getROIOrigin = self._roiControl.getROIOrigin
        self.getImgSubsampling = self._imgOpts.getSubsampling
        self.getMode = self._mainWidget.getMode
        self.sigModeChanged = self._mainWidget.sigModeChanged
        self.isModeLock = self._mainWidget.isModeLock
        self.setModeLock = self._mainWidget.setModeLock

        # set up interface
        self.setAxis(self._recons_params)

    def setScan(self, scan):
        self._mainWidget.setScan(scan=scan)
        self._roiControl.setScan(scan=scan)

    def manual_uses_full_image(self, value):
        self._roiControl.manual_uses_full_image(value)

    def _urlChanged(self):
        self.sigUrlChanged.emit()

    def _subsamplingChanged(self):
        self.sigSubsamplingChanged.emit()

    def _incrementLeftShift(self):
        self._incrementShift("left")

    def _incrementRightShift(self):
        self._incrementShift("right")

    def _incrementTopShift(self):
        self._incrementShift("top")

    def _incrementBottomShift(self):
        self._incrementShift("bottom")

    def _setShiftAndSignal(self, x, y):
        if x == self._xShift and y == self._yShift:
            return
        self.setShift(x, y)
        self._shiftControl._updateShiftInfo(x=x, y=y)
        self.sigShiftChanged.emit(x, y)

    def setAxis(self, axis):
        if axis == self._axis:
            return
        assert isinstance(axis, QAxisRP)
        old = self.blockSignals(True)
        if self._axis:
            self._axis.sigChanged.disconnect(self._updateAxisView)
        self._axis = axis
        self.setXShift(self._axis.relative_cor_value)
        self._mainWidget.setAxisParams(self._axis)
        self._readFileSelWidget.setAxis(self._axis)
        self._updateAxisView()
        self._axis.sigChanged.connect(self._updateAxisView)
        self.blockSignals(old)

    def _modeLockChanged(self, lock):
        self.sigModeLockChanged.emit(lock)

    def setMode(self, mode):
        old = self._axis.blockSignals(True)
        self._axis.mode = mode
        self._mainWidget._calculationWidget.setMode(mode)
        self._updateAxisView()
        self._axis.blockSignals(old)
        self._axis.sigChanged.emit()

    def setEstimatedCor(self, value):
        self._mainWidget.setEstimatedCorValue(value=value)

    def getEstimatedCor(self):
        return self._mainWidget.getEstimatedCor()

    def _updateAxisView(self):
        old = self._axis.blockSignals(True)
        if self._axis.relative_cor_value not in (None, "..."):
            self.setXShift(self._axis.relative_cor_value)
        self._axis.blockSignals(old)

        self._manualSelectionWidget.setVisible(
            self._axis.mode is axis_mode.AxisMode.manual
        )
        self._readFileSelWidget.setVisible(self._axis.mode is axis_mode.AxisMode.read)

    def getAxis(self):
        return self._axis

    def _incrementShift(self, direction):
        assert direction in ("left", "right", "top", "bottom")
        if direction == "left":
            self.setXShift(self._xShift - self.getShiftStep())
        elif direction == "right":
            self.setXShift(self._xShift + self.getShiftStep())
        elif direction == "top":
            self.setYShift(self._yShift + self.getShiftStep())
        else:
            self.setYShift(self._yShift - self.getShiftStep())

        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)

    def _resetShift(self):
        old = self._axis.blockSignals(True)
        self.setXShift(0)
        self.setYShift(0)
        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)
        self._axis.blockSignals(old)

        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def getXShift(self):
        if self._xShift == "...":
            return 0
        return self._xShift

    def getYShift(self):
        if self._yShift == "...":
            return 0
        return self._yShift

    def setXShift(self, x: float):
        self.setShift(x=x, y=self._yShift)

    def setYShift(self, y):
        self.setShift(x=self._xShift, y=y)

    def setShift(self, x, y):
        if x == self._xShift and y == self._yShift:
            return
        self._xShift = x if x is not None else 0.0
        self._yShift = y if y is not None else 0.0
        if self._axis:
            old = self._axis.blockSignals(True)
            self._axis.set_relative_value(x)
            self._axis.blockSignals(old)
        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)
        if not isinstance(self._xShift, str):
            # filter `...` and `?` values (used for issues or processing)
            self.sigShiftChanged.emit(self._xShift, self._yShift)

    def reset(self):
        old = self.blockSignals(True)
        self.setShift(0, 0)
        self.blockSignals(old)
        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def setLocked(self, locked):
        self._mainWidget.setEnabled(not locked)

    def _requestZoomReset(self):
        self.sigResetZoomRequested.emit()

    def getROIDims(self):
        return self._roiControl.getROIDims()


class _AxisManualSelection(qt.QWidget):

    sigResetZoomRequested = qt.Signal()
    """Signal emitted when a zoom request is necessary (when change to full
    image)"""

    def __init__(self, parent, shift_mode):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._displacementSelector = _DisplacementSelector(parent=self)
        self.layout().addWidget(self._displacementSelector)

        self._shiftControl = _ShiftControl(parent=self, shift_mode=shift_mode)
        self.layout().addWidget(self._shiftControl)

        self._roiControl = _ROIControl(parent=self)
        self.layout().addWidget(self._roiControl)

        self._imgOpts = _ImgOpts(parent=self)
        self.layout().addWidget(self._imgOpts)

        # connect signal / slot
        self._roiControl.sigResetZoomRequested.connect(self._requestZoomReset)

    def _requestZoomReset(self):
        self.sigResetZoomRequested.emit()


class _ROIControl(qt.QGroupBox):
    """
    Widget used to define the ROI on images to compare
    """

    sigRoiChanged = qt.Signal(object)
    """Signal emitted when the ROI changed"""
    sigResetZoomRequested = qt.Signal()
    """Signal emitted when a zoom request is necessary (when change to full
    image)"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "ROI selection", parent)
        self.setLayout(qt.QVBoxLayout())

        self._buttonGrp = qt.QButtonGroup(parent=self)
        self._buttonGrp.setExclusive(True)

        self._roiWidget = qt.QWidget(parent=self)
        self._roiWidget.setLayout(qt.QHBoxLayout())
        self._roiWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._fullImgButton = qt.QRadioButton("full image", parent=self)
        self._buttonGrp.addButton(self._fullImgButton)
        self.layout().addWidget(self._fullImgButton)
        self._roiButton = qt.QRadioButton("ROI", parent=self._roiWidget)
        self._roiWidget.layout().addWidget(self._roiButton)
        self._buttonGrp.addButton(self._roiButton)
        self._roiDefinition = _ROIDefinition(parent=self)
        self._roiWidget.layout().addWidget(self._roiDefinition)
        self.layout().addWidget(self._roiWidget)

        # connect signal / Slot
        self._roiButton.toggled.connect(self._roiDefinition.setEnabled)
        self._fullImgButton.toggled.connect(self._requestZoomReset)
        self._roiButton.toggled.connect(self._requestZoomReset)

        # expose API
        self.sigRoiChanged = self._roiDefinition.sigRoiChanged
        self.getROIOrigin = self._roiDefinition.getROIOrigin
        self.setLimits = self._roiDefinition.setLimits
        self.setScan = self._roiDefinition.setScan

        # setup for full image
        self._fullImgButton.setChecked(True)

    def getROIDims(self):
        if self._roiButton.isChecked():
            return self._roiDefinition.getROIDims()
        else:
            return None

    def manual_uses_full_image(self, activate):
        if activate:
            self._fullImgButton.setChecked(True)
        else:
            self._roiButton.setChecked(True)

    def _requestZoomReset(self):
        self.sigResetZoomRequested.emit()


class _ROIDefinition(qt.QWidget):
    """
    Widget used to define ROI width and height.

    :note: emit ROI == None if setDisabled
    """

    sigRoiChanged = qt.Signal(object)
    """Signal emitted when the ROI changed"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self._already_set = False

        # width & height
        self.layout().addWidget(qt.QLabel("dims", self), 0, 0)
        self._widthSB = qt.QSpinBox(parent=self)
        self._widthSB.setSingleStep(2)
        self._widthSB.setMaximum(10000)
        self._widthSB.setSuffix(" px")
        self._widthSB.setPrefix("w: ")
        self._widthSB.setToolTip("ROI width")
        self.layout().addWidget(self._widthSB, 0, 1)
        self._heightSB = qt.QSpinBox(parent=self)
        self._heightSB.setSingleStep(2)
        self._heightSB.setSuffix(" px")
        self._heightSB.setPrefix("h: ")
        self._heightSB.setToolTip("ROI height")
        self._heightSB.setMaximum(10000)
        self.layout().addWidget(self._heightSB, 0, 2)

        # origin x and y position
        self.layout().addWidget(qt.QLabel("origin", self), 1, 0)
        self._xOriginSB = qt.QSpinBox(parent=self)
        self._xOriginSB.setSingleStep(10)
        self._xOriginSB.setMaximum(10000)
        self._xOriginSB.setPrefix("x: ")
        self.layout().addWidget(self._xOriginSB, 1, 1)
        self._yOriginSB = qt.QSpinBox(parent=self)
        self._yOriginSB.setSingleStep(10)
        self._yOriginSB.setPrefix("y: ")
        self._yOriginSB.setMaximum(10000)
        self.layout().addWidget(self._yOriginSB, 1, 2)

        # Signal / Slot connection
        self._widthSB.editingFinished.connect(self.__roiChanged)
        self._heightSB.editingFinished.connect(self.__roiChanged)
        self._xOriginSB.editingFinished.connect(self.__roiChanged)
        self._yOriginSB.editingFinished.connect(self.__roiChanged)

    def __roiChanged(self, *args, **kwargs):
        self.sigRoiChanged.emit((self.getROIDims(), self.getROIOrigin()))

    def setLimits(self, width, height):
        """

        :param int x: width maximum value
        :param int height: height maximum value
        """
        for spinButton in (self._widthSB, self._heightSB):
            spinButton.blockSignals(True)
        assert type(width) is int
        assert type(height) is int
        valueChanged = False
        if self._widthSB.value() > width:
            self._widthSB.setValue(width)
            valueChanged = True
        if self._heightSB.value() > height:
            self._heightSB.setValue(height)
            valueChanged = True

        # if this is the first limit definition, propose default width and
        # height
        if self._widthSB.value() == 0:
            self._widthSB.setValue(min(256, width))
            valueChanged = True
        if self._heightSB.value() == 0:
            self._heightSB.setValue(min(256, height))
            valueChanged = True

        # define minimum / maximum
        self._widthSB.setRange(1, width)
        self._heightSB.setRange(1, height)
        for spinButton in (self._widthSB, self._heightSB):
            spinButton.blockSignals(False)
        if valueChanged is True:
            self.__roiChanged()

    def getROIDims(self):
        """

        :return: (width, height) or None
        :rtype: Union[None, tuple]
        """
        if self.isEnabled():
            return (self._widthSB.value(), self._heightSB.value())
        else:
            return None

    def getROIOrigin(self):
        return (self._xOriginSB.value(), self._yOriginSB.value())

    def setEnabled(self, *arg, **kwargs):
        qt.QWidget.setEnabled(self, *arg, **kwargs)
        self.__roiChanged()

    def setScan(self, scan):
        if not self._already_set:
            self._already_set = True
            try:
                x_origin = scan.dim_1 // 2
                y_origin = scan.dim_2 // 2
                self._xOriginSB.setValue(x_origin)
                self._yOriginSB.setValue(y_origin)
            except Exception:
                _logger.warning("unable to determine origin for {}".format(str(scan)))


@enum.unique
class ShiftMode(enum.Enum):
    x_only = 0
    y_only = 1
    x_and_y = 2


class _ShiftControl(qt.QWidget):
    """
    Widget to control the shift step we want to apply
    """

    sigShiftLeft = qt.Signal()
    """Signal emitted when the left button is activated"""
    sigShiftRight = qt.Signal()
    """Signal emitted when the right button is activated"""
    sigShiftTop = qt.Signal()
    """Signal emitted when the top button is activated"""
    sigShiftBottom = qt.Signal()
    """Signal emitted when the bottom button is activated"""
    sigReset = qt.Signal()
    """Signal emitted when the reset button is activated"""
    sigAuto = qt.Signal()
    """Signal emitted when the auto button is activated"""
    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted ony when xLE and yLE edition is finished"""

    def __init__(self, parent, shift_mode):
        """

        :param parent: qt.QWidget
        :param ShiftMode shift_mode: what are the shift we want to control
        """
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._leftButton = qt.QPushButton("left", parent=self)
        self.layout().addWidget(self._leftButton, 1, 0)

        self._rightButton = qt.QPushButton("right", parent=self)
        self.layout().addWidget(self._rightButton, 1, 3)

        self._shiftInfo = _ShiftInformation(parent=self)
        self.layout().addWidget(self._shiftInfo, 1, 1)
        self._shiftInfo._updateShiftInfo(x=0.0, y=0.0)

        self._topButton = qt.QPushButton("top", parent=self)
        self.layout().addWidget(self._topButton, 0, 1)

        self._bottomButton = qt.QPushButton("bottom", parent=self)
        self.layout().addWidget(self._bottomButton, 2, 1)

        self._resetButton = qt.QPushButton("reset", parent=self)
        self.layout().addWidget(self._resetButton, 3, 2, 3, 4)

        self._autoButton = qt.QPushButton("auto", parent=self)
        self.layout().addWidget(self._autoButton, 3, 0, 3, 2)
        self._autoButton.hide()

        # Signal / Slot connection
        self._leftButton.pressed.connect(self.sigShiftLeft.emit)
        self._rightButton.pressed.connect(self.sigShiftRight.emit)
        self._topButton.pressed.connect(self.sigShiftTop.emit)
        self._bottomButton.pressed.connect(self.sigShiftBottom.emit)
        self._resetButton.pressed.connect(self.sigReset.emit)
        self._autoButton.pressed.connect(self.sigAuto.emit)
        self._shiftInfo.sigShiftChanged.connect(self.sigShiftChanged.emit)

        # expose API
        self._updateShiftInfo = self._shiftInfo._updateShiftInfo

        self.setShiftMode(shift_mode)

    def setShiftMode(self, shift_mode):
        show_x_shift = shift_mode in (ShiftMode.x_only, ShiftMode.x_and_y)
        show_y_shift = shift_mode in (ShiftMode.y_only, ShiftMode.x_and_y)
        self._leftButton.setVisible(show_x_shift)
        self._rightButton.setVisible(show_x_shift)
        self._topButton.setVisible(show_y_shift)
        self._bottomButton.setVisible(show_y_shift)
        self._shiftInfo._xLE.setVisible(show_x_shift)
        self._shiftInfo._xLabel.setVisible(show_x_shift)
        self._shiftInfo._yLE.setVisible(show_y_shift)
        self._shiftInfo._yLabel.setVisible(show_y_shift)


class _ImgOpts(qt.QGroupBox):

    sigSubsamplingChanged = qt.Signal()
    """Signal emitted when the subsampling change"""

    def __init__(self, parent, title="Image Option"):
        super().__init__(title, parent)
        self.setLayout(qt.QFormLayout())
        self._subsamplingQSpinBox = qt.QSpinBox(self)
        self.layout().addRow("subsampling:", self._subsamplingQSpinBox)
        self._subsamplingQSpinBox.setMinimum(1)

        # set up
        self._subsamplingQSpinBox.setValue(1)

        # connect signal / slot
        self._subsamplingQSpinBox.valueChanged.connect(self._subsamplingChanged)

    def _subsamplingChanged(self):
        self.sigSubsamplingChanged.emit()

    def getSubsampling(self):
        return self._subsamplingQSpinBox.value()

    def setSubsampling(self, value):
        return self._subsamplingQSpinBox.setValue(int(value))


class _ShiftInformation(qt.QWidget):
    """
    Widget displaying information about the current x and y shift.
    Both x shift and y shift are editable.
    """

    class _ShiftLineEdit(qt.QLineEdit):
        def __init__(self, *args, **kwargs):
            qt.QLineEdit.__init__(self, *args, **kwargs)
            validator = qt.QDoubleValidator(parent=self, decimals=2)
            self.setValidator(validator)

        def sizeHint(self):
            return qt.QSize(40, 10)

    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted ony when xLE and yLE edition is finished"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self._xLabel = qt.QLabel("x=", parent=self)
        self.layout().addWidget(self._xLabel)
        self._xLE = _ShiftInformation._ShiftLineEdit("", parent=self)
        self.layout().addWidget(self._xLE)

        self._yLabel = qt.QLabel("y=", parent=self)
        self.layout().addWidget(self._yLabel)
        self._yLE = _ShiftInformation._ShiftLineEdit("", parent=self)
        self.layout().addWidget(self._yLE)

        # connect Signal / Slot
        self._xLE.editingFinished.connect(self._shiftChanged)
        self._yLE.editingFinished.connect(self._shiftChanged)

    def _updateShiftInfo(self, x, y):
        self.blockSignals(True)
        if x is None:
            x = 0.0
        if y is None:
            y = 0.0
        x_text = x
        if x_text != "...":
            x_text = "%.1f" % float(x)
        y_text = y
        if y_text != "...":
            y_text = "%.1f" % float(y)
        self._xLE.setText(x_text)
        self._yLE.setText(y_text)
        self.blockSignals(False)

    def _shiftChanged(self, *args, **kwargs):
        self.sigShiftChanged.emit(float(self._xLE.text()), float(self._yLE.text()))


class _DisplacementSelector(qt.QGroupBox):
    """
    Group box to define the displacement step value
    """

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "shift step", parent)
        self.setLayout(qt.QVBoxLayout())
        self._buttonGrp = qt.QButtonGroup(parent=self)
        self._buttonGrp.setExclusive(True)

        self._rawButton = qt.QRadioButton("Raw (1 pixel)", parent=self)
        self.layout().addWidget(self._rawButton)
        self._buttonGrp.addButton(self._rawButton)

        self._fineButton = qt.QRadioButton("Fine (0.1 pixel)", parent=self)
        self.layout().addWidget(self._fineButton)
        self._buttonGrp.addButton(self._fineButton)

        self._manualWidget = qt.QWidget(parent=self)
        self._manualWidget.setLayout(qt.QHBoxLayout())
        self._manualWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._manualWidget.layout().setSpacing(0)
        self._manualButton = qt.QRadioButton("Manual", parent=self._manualWidget)
        self._manualWidget.layout().addWidget(self._manualButton)
        self._manualLE = qt.QLineEdit("0.5", parent=self._manualWidget)
        validator = qt.QDoubleValidator(parent=self._manualLE, decimals=2)
        validator.setBottom(0.0)
        self._manualLE.setValidator(validator)
        self._manualWidget.layout().addWidget(self._manualLE)

        self.layout().addWidget(self._manualWidget)
        self._manualLE.setEnabled(False)
        self._buttonGrp.addButton(self._manualButton)

        self._rawButton.setChecked(True)

        # signal / slot connection
        self._manualButton.toggled.connect(self._manualLE.setEnabled)

    def getShiftStep(self):
        """

        :return: displacement shift defined
        :rtype: float
        """
        if self._rawButton.isChecked():
            return 1.0
        elif self._fineButton.isChecked():
            return 0.1
        else:
            return float(self._manualLE.text())

    def setShiftStep(self, value):
        """

        :param float value: shift step
        """
        assert type(value) is float
        self._manualButton.setChecked(True)
        self._manualLE.setText(str(value))


class _AxisOptionsWidget(qt.QWidget):
    """GUI to tune the axis algorithm"""

    def __init__(self, parent, axis):
        qt.QWidget.__init__(self, parent=parent)
        assert isinstance(axis, QAxisRP)
        self._axis = axis
        self.setLayout(qt.QVBoxLayout())

        # define common options
        self._commonOpts = qt.QWidget(parent=self)
        self._commonOpts.setLayout(qt.QFormLayout())

        self._qcbDataMode = qt.QComboBox(parent=self)
        for data_mode in AxisCalculationInput:
            # paganin is not managed for sinogram
            self._qcbDataMode.addItem(data_mode.name(), data_mode)
        # for now not handle
        # self._commonOpts.layout().addRow('data mode', self._qcbDataMode)
        self._qcbDataMode.hide()

        # add scale option
        self._scaleOpt = qt.QCheckBox(parent=self)
        self._commonOpts.layout().addRow("scale the two images", self._scaleOpt)
        self.layout().addWidget(self._commonOpts)

        # add option for computing min-max
        # TODO
        pass

        # add near options
        self._nearOpts = _AxisNearOptsWidget(parent=self, axis=self._axis)
        self.layout().addWidget(self._nearOpts)

        # set up
        self.setCalculationInputType(self._axis.calculation_input_type)

        # connect signal / slot
        self._scaleOpt.toggled.connect(self._updateScaleOpt)
        self._qcbDataMode.currentIndexChanged.connect(self._updateInputType)
        self._axis.sigChanged.connect(self._updateMode)

    def setMode(self, mode):
        mode_ = axis_mode.AxisMode.from_value(mode)
        self._nearOpts.setVisible(mode_ == axis_mode.AxisMode.near)

    def _updateMode(self):
        old = self.blockSignals(True)
        self._nearOpts.setVisible(self._axis.mode == axis_mode.AxisMode.near)
        index = self._qcbDataMode.findText(self._axis.calculation_input_type.name())
        if index >= 0:
            self._qcbDataMode.setCurrentIndex(index)
        self.blockSignals(old)

    def _updateScaleOpt(self, *arg, **kwargs):
        self._axis.scale_img2_to_img1 = self.isImageScaled()

    def isImageScaled(self):
        return self._scaleOpt.isChecked()

    def _updateInputType(self, *arg, **kwargs):
        self._axis.calculation_input_type = self.getCalulationInputType()

    def getCalulationInputType(self, *arg, **kwargs):
        return AxisCalculationInput.from_value(self._qcbDataMode.currentText())

    def setCalculationInputType(self, calculation_type):
        calculation_type = AxisCalculationInput.from_value(calculation_type)
        index_dm = self._qcbDataMode.findText(calculation_type.name())
        self._qcbDataMode.setCurrentIndex(index_dm)

    def setAxisParams(self, axis):
        self._nearOpts.setAxisParams(axis=axis)
        self._axis = axis
        old = self.blockSignals(True)
        self._scaleOpt.setChecked(axis.scale_img2_to_img1)
        index = self._qcbDataMode.findText(axis.calculation_input_type.name())
        self._qcbDataMode.setCurrentIndex(index)
        self.blockSignals(old)


class _AxisNearOptsWidget(qt.QWidget):
    """GUI dedicated to the neat option"""

    def __init__(self, parent, axis):
        qt.QWidget.__init__(self, parent=parent)
        assert isinstance(axis, QAxisRP)
        self._axis = axis

        self.setLayout(qt.QFormLayout())

        self._stdMaxOpt = qt.QCheckBox(parent=self)
        self.layout().addRow("look at max standard deviation", self._stdMaxOpt)

        self._nearWX = qt.QSpinBox(parent=self)
        self._nearWX.setMinimum(1)
        self._nearWX.setValue(5)
        self.layout().addRow("window size", self._nearWX)

        self._fineStepX = qt.QDoubleSpinBox(parent=self)
        self._fineStepX.setMinimum(0.05)
        self._fineStepX.setSingleStep(0.05)
        self._fineStepX.setMaximum(1.0)
        self.layout().addRow("fine step x", self._fineStepX)

        # connect signal / Slot
        self._stdMaxOpt.toggled.connect(self._lookforStxMaxChanged)
        self._nearWX.valueChanged.connect(self._windowSizeChanged)
        self._fineStepX.valueChanged.connect(self._fineStepXChanged)

    def _lookforStxMaxChanged(self, *args, **kwargs):
        self._axis.look_at_stdmax = self.isLookAtStdMax()

    def isLookAtStdMax(self):
        """

        :return: is the option for looking at max standard deviation is
                 activated
        :rtype: bool
        """
        return self._stdMaxOpt.isChecked()

    def _windowSizeChanged(self, *args, **kwargs):
        self._axis.near_wx = self.getWindowSize()

    def getWindowSize(self):
        """

        :return: window size for near search
        :rtype: int
        """
        return self._nearWX.value()

    def _fineStepXChanged(self, *args, **kwargs):
        self._axis.fine_step_x = self.getFineStepX()

    def getFineStepX(self):
        """

        :return: fine step x for near calculation
        :rtype: float
        """
        return self._fineStepX.value()

    def setAxisParams(self, axis):
        """

        :param axis: axis to edit
        :type: AxisRP
        """
        old = self.blockSignals(True)
        self._axis = axis
        self._stdMaxOpt.setChecked(axis.look_at_stdmax)
        self._nearWX.setValue(axis.near_wx)
        self._fineStepX.setValue(axis.fine_step_x)
        self.blockSignals(old)


class AxisTabWidget(qt.QTabWidget):
    """
    TabWidget containing all the information to edit the AXIS parameters
    """

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the mode lock change"""

    sigUrlChanged = qt.Signal()
    """Signal emit when frames urls changed"""

    def __init__(
        self,
        recons_params,
        parent=None,
        mode_dependant_widget=None,
        read_file_sel_widget=None,
    ):
        """

        :param recons_params: reconstruction parameters edited by the widget
        :type: QAxisRP
        :param mode_dependant_widget: widget used for manual selection of the
                                      axis
        :type mode_dependant_widget: Union[None, `._AxisManualSelection`]
        :param read_file_sel_widget: widget used to select a par file containing
                                     the axis position
        :type read_file_sel_widget: Union[None, `._AxisRead`]
        """
        qt.QTabWidget.__init__(self, parent)
        assert recons_params is not None
        # first tab 'calculation widget'
        self._calculationWidget = _CalculationWidget(
            parent=self, axis_params=recons_params
        )

        # second tab: options
        self._optionsWidget = _AxisOptionsWidget(parent=self, axis=recons_params)
        self._inputWidget = _InputWidget(parent=self, axis_params=recons_params)

        if mode_dependant_widget:
            self._calculationWidget.layout().addWidget(mode_dependant_widget)

        if read_file_sel_widget:
            self._calculationWidget.layout().addWidget(read_file_sel_widget)

        for widget in self._calculationWidget, self._optionsWidget:
            spacer = qt.QWidget(self)
            spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
            widget.layout().addWidget(spacer)

        self._optionsSA = qt.QScrollArea(parent=self)
        self._optionsSA.setWidget(self._optionsWidget)
        self.addTab(self._calculationWidget, "calculation")
        self.addTab(self._optionsSA, "options")
        # simplify set up. Hide options
        self._optionsSA.hide()
        self.addTab(self._inputWidget, "input")

        # set up
        self.setAxisParams(recons_params)
        self._optionsWidget.setMode(self.getMode())
        self._updatePossibleInput()

        # expose API
        self.sigModeChanged = self._calculationWidget.sigModeChanged
        self.isModeLock = self._calculationWidget.isModeLock
        self.setModeLock = self._calculationWidget.setModeLock
        self.setEstimatedCorValue = self._calculationWidget.setEstimatedCorValue
        self.getEstimatedCor = self._calculationWidget.getEstimatedCor

        # connect signal / slot
        self._calculationWidget.sigLockModeChanged.connect(
            self._propagateSigLockModeChanged
        )
        self.sigModeChanged.connect(self._updatePossibleInput)
        self._inputWidget._sigUrlChanged.connect(self._urlChanged)

    def _urlChanged(self):
        self.sigUrlChanged.emit()

    def _propagateSigLockModeChanged(self, lock):
        self.sigLockModeChanged.emit(lock)

    def getMode(self):
        """Return algorithm to use for axis calculation"""
        return self._calculationWidget.getMode()

    def setScan(self, scan):
        if scan is not None:
            self._inputWidget.setScan(scan)

    def setAxisParams(self, axis):
        old = self.blockSignals(True)
        self._calculationWidget.setAxisParams(axis)
        self._optionsWidget.setAxisParams(axis)
        self._inputWidget.setAxisParams(axis)
        self.blockSignals(old)

    def _updatePossibleInput(self):
        """Update Input tab according to the current mode"""
        current_mode = self.getMode()
        if current_mode not in axis_mode._VALID_INPUTS:
            raise ValueError(
                "valid input should have been defined for the "
                "mode {}".format(current_mode.value)
            )
        else:
            valid_inputs = axis_mode._VALID_INPUTS[current_mode]
            if valid_inputs is None:
                self._inputWidget.setEnabled(False)
            else:
                self._inputWidget.setEnabled(True)
                self._inputWidget.setValidInputs(valid_inputs)


class _CalculationWidget(qt.QWidget):
    """Main widget to select the algorithm to use for COR calculation"""

    sigModeChanged = qt.Signal(str)
    """signal emitted when the algorithm for computing COR changed"""

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the mode has been lock or unlock"""

    def __init__(self, parent, axis_params):
        assert isinstance(axis_params, QAxisRP)
        qt.QWidget.__init__(self, parent)
        self._axis_params = None
        self.setLayout(qt.QVBoxLayout())

        self._modeWidget = qt.QWidget(parent=self)
        self._modeWidget.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._modeWidget)

        self.__rotAxisSelLabel = qt.QLabel("algorithm to compute cor")
        self._modeWidget.layout().addWidget(self.__rotAxisSelLabel)
        self._qcbPosition = qt.QComboBox(self)
        # add centered value
        self._qcbPosition.addItem(axis_mode.AxisMode.centered.value)
        idx = self._qcbPosition.findText(axis_mode.AxisMode.centered.value)
        global_tooltip = 'Dedicated to fullfield. Previously named "accurate"'
        self._qcbPosition.setItemData(idx, global_tooltip, qt.Qt.ToolTipRole)
        # add global value
        self._qcbPosition.addItem(axis_mode.AxisMode.global_.value)
        idx = self._qcbPosition.findText(axis_mode.AxisMode.global_.value)
        global_tooltip = (
            "Algorithm which can work for both half acquisition"
            " and standard ('full field') acquisition"
        )
        self._qcbPosition.setItemData(idx, global_tooltip, qt.Qt.ToolTipRole)
        # add sliding window value
        self._qcbPosition.addItem(axis_mode.AxisMode.sliding_window.value)
        idx = self._qcbPosition.findText(axis_mode.AxisMode.sliding_window.value)
        sliding_w_tooltip = (
            "A method for estimating semi-automatically the CoR position. "
            "You have to provide a hint on where the CoR is "
            "(left, center, right). Also works for sinograms."
        )
        self._qcbPosition.setItemData(idx, sliding_w_tooltip, qt.Qt.ToolTipRole)
        # add growing window value
        self._qcbPosition.addItem(axis_mode.AxisMode.growing_window.value)
        idx = self._qcbPosition.findText(axis_mode.AxisMode.growing_window.value)
        growing_w_tooltip = "A auto-Cor method also working for sinograms."
        self._qcbPosition.setItemData(idx, growing_w_tooltip, qt.Qt.ToolTipRole)
        # add sino-coarse-to-fine value
        self._qcbPosition.addItem(axis_mode.AxisMode.sino_coarse_to_fine.value)
        idx = self._qcbPosition.findText(axis_mode.AxisMode.sino_coarse_to_fine.value)
        sino_coarse_to_fine_tooltip = (
            "Estimate CoR from sinogram. Only works for 360 degrees scans."
        )
        self._qcbPosition.setItemData(
            idx, sino_coarse_to_fine_tooltip, qt.Qt.ToolTipRole
        )
        # add composite coarse to fine
        self._qcbPosition.addItem(axis_mode.AxisMode.composite_coarse_to_fine.value)
        idx = self._qcbPosition.findText(
            axis_mode.AxisMode.composite_coarse_to_fine.value
        )
        composite_coarse_to_fine_tooltip = "Estimate CoR from composite multi-angle images. Only works for 360 degrees scans."
        self._qcbPosition.setItemData(
            idx, composite_coarse_to_fine_tooltip, qt.Qt.ToolTipRole
        )
        # add manual
        self._qcbPosition.addItem(axis_mode.AxisMode.manual.value)
        # add read value
        self._qcbPosition.addItem(axis_mode.AxisMode.read.value)

        # add tooltips
        def add_tooltip(value, tooltip):
            idx = self._qcbPosition.findText(value)
            self._qcbPosition.setItemData(idx, tooltip, qt.Qt.ToolTipRole)

        # centered tooltip
        tooltip = "Default algorithm to compute COR from full field acquisition"
        add_tooltip(value=axis_mode.AxisMode.centered.value, tooltip=tooltip)
        # near tooltip
        tooltip = "Default algorithm to compute COR from half field acquisition"
        add_tooltip(value=axis_mode.AxisMode.near.value, tooltip=tooltip)
        # manual tooltip
        tooltip = "Enter manually the COR value"
        add_tooltip(value=axis_mode.AxisMode.manual.value, tooltip=tooltip)
        # read tooltip
        tooltip = "Read COR value from a file"
        add_tooltip(value=axis_mode.AxisMode.read.value, tooltip=tooltip)

        self._modeWidget.layout().addWidget(self._qcbPosition)

        # method lock button
        self._lockMethodPB = PadlockButton(parent=self._modeWidget)
        self._lockMethodPB.setToolTip(
            "Lock the method to compute the cor. \n"
            "This will automatically call the "
            "defined algorithm each time a scan is received."
        )
        self._modeWidget.layout().addWidget(self._lockMethodPB)

        self._optsWidget = qt.QWidget(self)
        self._optsWidget.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self._optsWidget)

        # padding option
        self._padding_widget = qt.QGroupBox("padding mode")
        self._padding_widget.setCheckable(True)
        self._optsWidget.layout().addWidget(self._padding_widget)
        self._padding_widget.setLayout(qt.QHBoxLayout())

        self._qbPaddingMode = qt.QComboBox(self._padding_widget)
        for _mode in (
            "constant",
            "edge",
            "linear_ramp",
            "maximum",
            "mean",
            "median",
            "minimum",
            "reflect",
            "symmetric",
            "wrap",
        ):
            self._qbPaddingMode.addItem(_mode)
        def_index = self._qbPaddingMode.findText("edge")
        self._qbPaddingMode.setCurrentIndex(def_index)
        self._padding_widget.layout().addWidget(self._qbPaddingMode)

        # side option
        self._sideWidget = qt.QWidget(self)
        self._sideWidget.setLayout(qt.QHBoxLayout())
        self._optsWidget.layout().addWidget(self._sideWidget)
        self._sideWidget.layout().addWidget(qt.QLabel("side:", self))
        self._sideCB = qt.QComboBox(self._optsWidget)
        self._sideWidget.layout().addWidget(self._sideCB)
        self._sideCB.setToolTip(
            "Define a side for the sliding and growing" "window algorithms"
        )

        # near mode options
        self._nearOptsWidget = qt.QWidget(parent=self)
        self._nearOptsWidget.setLayout(qt.QVBoxLayout())
        self._optsWidget.layout().addWidget(self._nearOptsWidget)

        #    near value lock button
        self._nearValueCB = qt.QCheckBox("Overwrite estimated value")
        self._nearValueCB.setToolTip(
            "If the acquisition contains an "
            "estimation of the COR value then "
            "will set it automatically as estimated "
            "value."
        )
        self._nearOptsWidget.layout().addWidget(self._nearValueCB)
        self._nearOptsWidget.hide()

        #    LineEdit position value
        self._qleValueW = qt.QWidget(self._nearOptsWidget)
        self._qleValueW.setLayout(qt.QFormLayout())
        self._nearOptsWidget.layout().addWidget(self._qleValueW)

        self._qleNearPosQLE = qt.QLineEdit("0", self._nearOptsWidget)
        validator = qt.QDoubleValidator(self._qleNearPosQLE)
        self._qleNearPosQLE.setValidator(validator)
        self._qleValueW.layout().addRow("estimated value:", self._qleNearPosQLE)

        # connect signal / slot
        self._qcbPosition.currentIndexChanged.connect(self._modeChanged)
        self._qleNearPosQLE.editingFinished.connect(self._nearValueChanged)
        self._sideCB.currentTextChanged.connect(self._sideChanged)
        self._lockMethodPB.sigLockChanged.connect(self.lockMode)
        self._qbPaddingMode.currentIndexChanged.connect(self._paddingModeChanged)
        self._padding_widget.toggled.connect(self._paddingModeChanged)

        # set up interface
        self._sideWidget.setVisible(False)
        self.setAxisParams(axis_params)
        self._nearValueCB.setChecked(True)

    def getMethodLockPB(self) -> qt.QPushButton:
        return self._lockMethodPB

    def setEstimatedCorValue(self, value):
        if value is not None:
            self._qleNearPosQLE.setText(str(value))

    def getEstimatedCor(self):
        return float(self._qleNearPosQLE.text())

    def nearValueIsLocked(self):
        return not self._nearValueCB.isChecked()

    def setSide(self, side):
        if side is not None:
            idx = self._sideCB.findText(side)
            if idx >= 0:
                self._sideCB.setCurrentIndex(idx)

    def getSide(self):
        return self._sideCB.currentText()

    def _modeChanged(self, *args, **kwargs):
        mode = self.getMode()
        old_0 = self._qcbPosition.blockSignals(True)
        old = self._axis_params.blockSignals(True)
        self._qleNearPosQLE.setVisible(mode == axis_mode.AxisMode.near)
        self._padding_widget.setVisible(
            mode
            in (
                axis_mode.AxisMode.centered,
                axis_mode.AxisMode.global_,
                axis_mode.AxisMode.growing_window,
                axis_mode.AxisMode.sliding_window,
                axis_mode.AxisMode.sino_coarse_to_fine,
                axis_mode.AxisMode.composite_coarse_to_fine,
            )
        )
        if mode in (
            axis_mode.AxisMode.near,
            axis_mode.AxisMode.centered,
            axis_mode.AxisMode.global_,
            axis_mode.AxisMode.growing_window,
            axis_mode.AxisMode.sliding_window,
            axis_mode.AxisMode.sino_coarse_to_fine,
            axis_mode.AxisMode.composite_coarse_to_fine,
        ):
            self._lockMethodPB.setVisible(True)
        else:
            self._lockMethodPB.setVisible(False)
            self.lockMode(False)

        side_visible = mode in (
            axis_mode.AxisMode.growing_window,
            axis_mode.AxisMode.sliding_window,
        )
        self._sideWidget.setVisible(side_visible)
        if side_visible is True:
            self._updateSideVisible(mode)

        self._axis_params.mode = mode.value
        self._axis_params.blockSignals(old)
        self._axis_params.changed()
        self.sigModeChanged.emit(mode.value)
        self._qcbPosition.blockSignals(old_0)

    def _updateSideVisible(self, mode: axis_mode.AxisMode):
        if mode not in (
            axis_mode.AxisMode.growing_window,
            axis_mode.AxisMode.sliding_window,
        ):
            return
        else:
            current_value = self._axis_params.side
            old = self._sideCB.blockSignals(True)
            self._sideCB.clear()
            values = ["left", "right", "center"]
            if mode is axis_mode.AxisMode.growing_window:
                values.append("all")
            for value in values:
                self._sideCB.addItem(value)
            idx = self._sideCB.findText(current_value)
            if idx >= 0:
                self._sideCB.setCurrentIndex(idx)
            self._sideCB.blockSignals(old)
            self._axis_params.side = self.getSide()

    def isModeLock(self):
        return self._lockMethodPB.isLocked()

    def setModeLock(self, mode=None):
        """set a specific mode and lock it.

        :param mode: mode to lock. If None then keep the current mode
        """
        if mode is not None:
            mode = axis_mode.AxisMode.from_value(mode)
        if mode is None and self.getMode() not in (
            axis_mode.AxisMode.centered,
            axis_mode.AxisMode.near,
            axis_mode.AxisMode.global_,
            axis_mode.AxisMode.growing_window,
            axis_mode.AxisMode.sliding_window,
            axis_mode.AxisMode.sino_coarse_to_fine,
            axis_mode.AxisMode.composite_coarse_to_fine,
        ):
            raise ValueError(
                "Unable to lock the current mode is not an automatic algorithm"
            )
        elif mode != self.getMode() and mode not in (
            axis_mode.AxisMode.centered,
            axis_mode.AxisMode.near,
            axis_mode.AxisMode.global_,
            axis_mode.AxisMode.growing_window,
            axis_mode.AxisMode.sliding_window,
            axis_mode.AxisMode.sino_coarse_to_fine,
            axis_mode.AxisMode.composite_coarse_to_fine,
        ):
            raise ValueError("Unable to lock %s this is not a lockable mode")

        if mode is not None:
            self.setMode(mode)
        if not self._lockMethodPB.isLocked():
            old = self._lockMethodPB.blockSignals(True)
            self._lockMethodPB.setLock(True)
            self._lockMethodPB.blockSignals(old)
        self.lockMode(True)

    def lockMode(self, lock):
        old = self._lockMethodPB.blockSignals(True)
        self._lockMethodPB.setLock(lock)
        for widget in (self._qcbPosition, self._qleNearPosQLE):
            widget.setEnabled(not lock)
        self._lockMethodPB.blockSignals(old)
        self.sigLockModeChanged.emit(lock)

    def getMode(self):
        """Return algorithm to use for axis calculation"""
        return axis_mode.AxisMode.from_value(self._qcbPosition.currentText())

    def setMode(self, mode):
        old = self._qcbPosition.blockSignals(True)
        index = self._qcbPosition.findText(mode.value)
        if index >= 0:
            self._qcbPosition.setCurrentIndex(index)
        self._qcbPosition.blockSignals(old)

    def _nearValueChanged(self, *args, **kwargs):
        self._axis_params.estimated_cor = self.getEstimatedCor()

    @deprecated(replacement="getEstimatedCor", since_version="0.6")
    def getNearPosition(self):
        return self.getEstimatedCor()

    @deprecated(replacement="setEstimatedCorValue", since_version="0.6")
    def setNearPosition(self, position):
        self.setEstimatedCorValue(position)

    def _paddingModeChanged(self, *args, **kwargs):
        self._axis_params.padding_mode = self.getPaddingMode()

    def _sideChanged(self, *args, **kwargs):
        self._axis_params.side = self.getSide()

    def getPaddingMode(self):
        if self._padding_widget.isChecked():
            return self._qbPaddingMode.currentText()
        else:
            return None

    def setPaddingMode(self, mode):
        index = self._qbPaddingMode.findText(mode)
        if index >= 0:
            self._qbPaddingMode.setCurrentIndex(index)
        self._paddingModeChanged()

    def setAxisParams(self, axis):
        old = self.blockSignals(True)
        if self._axis_params is not None:
            self._axis_params.sigChanged.disconnect(self._axis_params_changed)
        self._axis_params = axis
        self._axis_params.sigChanged.connect(self._axis_params_changed)
        self._axis_params_changed()
        self.blockSignals(old)

    def _axis_params_changed(self, *args, **kwargs):
        self.setMode(self._axis_params.mode)
        self.setEstimatedCorValue(self._axis_params.estimated_cor)
        self.setSide(self._axis_params.side)
        if self._axis_params.mode in (
            axis_mode.AxisMode.growing_window,
            axis_mode.AxisMode.sliding_window,
        ):
            self._sideWidget.setVisible(True)
        else:
            self._sideWidget.setVisible(False)
        self._updateSideVisible(mode=self._axis_params.mode)
        self.setPaddingMode(self._axis_params.padding_mode)


class _SliceSelector(qt.QWidget):
    sigChanged = qt.Signal()
    """signal emit when the selected slice change"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._modeCB = qt.QComboBox(self)
        self._modeCB.addItem("middle")
        self._modeCB.addItem("other")
        self.layout().addWidget(self._modeCB)
        self._otherSB = qt.QSpinBox(self)
        self._otherSB.setRange(0, 10000)
        self.layout().addWidget(self._otherSB)

        # connect signal / slot
        self._otherSB.valueChanged.connect(self._valueChanged)
        self._modeCB.currentIndexChanged.connect(self._modeChanged)
        # set up
        self._modeChanged()

    def getSlice(self) -> Union[int, str]:
        "return a specific slice index or 'middle'"
        if self.getMode() == "middle":
            return "middle"
        else:
            return self._otherSB.value()

    def setSlice(self, slice):
        if slice is None:
            return
        if slice == "middle":
            idx = self._modeCB.findText("middle")
            self._modeCB.setCurrentIndex(idx)
        else:
            idx = self._modeCB.findText("other")
            self._modeCB.setCurrentIndex(idx)
            self._otherSB.setValue(slice)
        self.sigChanged.emit()

    def getMode(self):
        return self._modeCB.currentText()

    def _valueChanged(self):
        self.sigChanged.emit()

    def _modeChanged(self, *args, **kwargs):
        self._otherSB.setVisible(self.getMode() == "other")
        self._valueChanged()


class _InputWidget(qt.QWidget):
    """Widget used to define the radio to use for axis calculation from
    radios"""

    sigChanged = qt.Signal()
    """Signal emitted when input changed"""

    _sigUrlChanged = qt.Signal()
    """Signal emit when url to be used changed"""

    def __init__(self, parent=None, axis_params=None):
        assert isinstance(axis_params, QAxisRP)
        self._blockUpdateAxisParams = False
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        # radio input
        self._radioGB = qt.QGroupBox(self)
        self._radioGB.setTitle("radios")
        self._radioGB.setLayout(qt.QVBoxLayout())
        self._radioGB.setCheckable(True)
        self.layout().addWidget(self._radioGB)
        ## angle mode
        self._angleModeWidget = _AngleSelectionWidget(
            parent=self, axis_params=axis_params
        )
        self._radioGB.layout().addWidget(self._angleModeWidget)
        self._axis_params = axis_params

        # sinogram input
        self._sinogramGB = qt.QGroupBox(self)
        self._sinogramGB.setLayout(qt.QVBoxLayout())
        self._standardSinogramOpts = qt.QGroupBox(self)
        self._sinogramGB.layout().addWidget(self._standardSinogramOpts)
        self._standardSinogramOpts.setLayout(qt.QFormLayout())
        self._standardSinogramOpts.layout().setContentsMargins(0, 0, 0, 0)
        self._standardSinogramOpts.setTitle("standard options")

        self._sinogramGB.setTitle("sinogram")
        self._sinogramGB.setCheckable(True)
        self.layout().addWidget(self._sinogramGB)
        ##  sinogram line
        self._sinogramLineSB = _SliceSelector(self)
        self._standardSinogramOpts.layout().addRow("line", self._sinogramLineSB)
        ##  sinogram subsampling
        self._sinogramSubsampling = qt.QSpinBox(self)
        self._sinogramSubsampling.setRange(1, 1000)
        self._sinogramSubsampling.setValue(10)
        self._standardSinogramOpts.layout().addRow(
            "subsampling", self._sinogramSubsampling
        )

        self._spacer = qt.QWidget(self)
        self._spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self._spacer)

        ## options for the composite mode
        self._compositeOpts = qt.QGroupBox(self)
        self._compositeOpts.setTitle("composite options")
        self._sinogramGB.layout().addWidget(self._compositeOpts)
        self._compositeOpts.setLayout(qt.QFormLayout())
        self._compositeOpts.layout().setContentsMargins(0, 0, 0, 0)
        self._thetaSB = qt.QSpinBox(self)
        self._thetaSB.setRange(0, 360)
        self._thetaSB.setValue(DEFAULT_CMP_THETA)
        self._thetaSB.setToolTip("a radio will be picked each theta degres")
        self._thetaLabel = qt.QLabel("angle interval (in degree)", self)
        self._thetaLabel.setToolTip(
            "algorithm will take one projection each 'angle interval'. Also know as 'theta'"
        )
        self._compositeOpts.layout().addRow(self._thetaLabel, self._thetaSB)

        self._oversamplingSB = qt.QSpinBox(self)
        self._oversamplingSB.setValue(DEFAULT_CMP_OVERSAMPLING)
        self._oversamplingSB.setToolTip("sinogram oversampling")
        self._compositeOpts.layout().addRow("oversampling", self._oversamplingSB)

        self._subsamplingYSB = qt.QSpinBox(self)
        self._subsamplingYSB.setValue(DEFAULT_CMP_SUBSAMPLING_Y)
        self._subsamplingYSB.setToolTip("sinogram subsampling")
        self._compositeOpts.layout().addRow("subsampling y", self._subsamplingYSB)

        self._takeLogCB = qt.QCheckBox(self)
        self._takeLogCB.setToolTip("Take logarithm")
        self._takeLogCB.setChecked(DEFAULT_CMP_TAKE_LOG)
        self._takeTheLogLabel = qt.QLabel("linearisation (-log(I/I0))")
        self._takeTheLogLabel.setToolTip(
            "take (-log(I/I0)) as input. Also know as 'take_log' option"
        )
        self._compositeOpts.layout().addRow(self._takeTheLogLabel, self._takeLogCB)

        # set up
        self._sinogramGB.setChecked(False)

        # connect signal / slot
        self._sinogramGB.toggled.connect(self._sinogramChecked)
        self._radioGB.toggled.connect(self._radiosChecked)
        self._sinogramSubsampling.valueChanged.connect(self._changed)
        self._sinogramLineSB.sigChanged.connect(self._changed)
        self._thetaSB.valueChanged.connect(self._changed)
        self._oversamplingSB.valueChanged.connect(self._changed)
        self._subsamplingYSB.valueChanged.connect(self._changed)
        self._takeLogCB.toggled.connect(self._changed)
        self._angleModeWidget.sigChanged.connect(self._sigUrlChanged)

    def setScan(self, scan: TomwerScanBase):
        if scan is not None:
            self._angleModeWidget.setScan(scan)
            self._angleModeWidget.setScanRange(scan.scan_range)

    def setAxisParams(self, axis_params):
        o1 = axis_params.blockSignals(True)
        o2 = self._axis_params.blockSignals(True)
        self._blockUpdateAxisParams = True

        if axis_params is not None:
            assert isinstance(axis_params, QAxisRP)
            o4 = self._sinogramGB.blockSignals(True)
            self._sinogramChecked(axis_params.use_sinogram, on_load=True)
            self._sinogramGB.blockSignals(o4)
            self._sinogramLineSB.setSlice(axis_params.sinogram_line)
            self._sinogramSubsampling.setValue(axis_params.sinogram_subsampling)
            self.setCompositeOptions(axis_params.composite_options)
        self._angleModeWidget.setAxisParams(axis_params)
        self._axis_params = axis_params
        axis_params.blockSignals(o1)
        self._axis_params.blockSignals(o2)
        self._blockUpdateAxisParams = False

    def getSinogramLine(self) -> Union[str, int]:
        return self._sinogramLineSB.getSlice()

    def getSinogramSubsampling(self) -> int:
        return self._sinogramSubsampling.value()

    def _sinogramChecked(self, checked, on_load=False):
        old_r = self._radioGB.blockSignals(True)
        old_s = self._sinogramGB.blockSignals(True)
        if checked:
            self._radioGB.setChecked(False)
            self._sinogramGB.setChecked(True)
        elif self._radioGB.isEnabled():
            self._radioGB.setChecked(not checked)
        elif on_load:
            self._sinogramGB.setChecked(checked)
        else:
            # ignore it if radio disabled
            self._sinogramGB.setChecked(True)
        self._radioGB.blockSignals(old_r)
        self._sinogramGB.blockSignals(old_s)
        self._changed()

    def _radiosChecked(self, checked, on_load=False):
        old_r = self._radioGB.blockSignals(True)
        old_s = self._sinogramGB.blockSignals(True)
        if checked:
            self._sinogramGB.setChecked(False)
            self._radioGB.setChecked(True)
        elif self._sinogramGB.isEnabled():
            self._sinogramGB.setChecked(not checked)
        elif on_load:
            self._radioGB.setChecked(checked)
        else:
            # ignore it if sinogram disabled
            self._radioGB.setChecked(True)
        self._radioGB.blockSignals(old_r)
        self._sinogramGB.blockSignals(old_s)
        self._changed()

    def _changed(self, *args, **kwargs):
        self._updateAxisParams()
        self.sigChanged.emit()

    def _updateAxisParams(self):
        if not self._blockUpdateAxisParams:
            self._axis_params.use_sinogram = self._sinogramGB.isChecked()
            self._axis_params.sinogram_line = self.getSinogramLine()
            self._axis_params.sinogram_subsampling = self.getSinogramSubsampling()
            self._axis_params.composite_options = self.getCompositeOptions()

    def setValidInputs(self, modes: Union[list, tuple]):
        """
        Define possible inputs.

        :param Union[list,tuple] modes:
        :raises: ValueError if modes are invalid
        """
        modes = set(modes)
        for mode in modes:
            try:
                axis_mode._InputType.from_value(mode)
            except Exception:
                raise ValueError(
                    "mode {} should be an instance of {}".format(
                        mode, axis_mode._InputType
                    )
                )
        if len(modes) == 2:
            self._sinogramGB.setEnabled(True)
            self._radioGB.setEnabled(True)
        elif len(modes) > 2:
            raise ValueError("invalid input {}".format(modes))
        elif len(modes) < 0:
            raise ValueError("modes is empty")
        else:
            mode = axis_mode._InputType.from_value(modes.pop())
            if mode in (axis_mode._InputType.SINOGRAM, axis_mode._InputType.COMPOSITE):
                self._sinogramGB.setEnabled(True)
                self._radioGB.setEnabled(False)
                self._sinogramGB.setChecked(True)
                self._compositeOpts.setEnabled(mode is axis_mode._InputType.COMPOSITE)
                self._standardSinogramOpts.setEnabled(
                    mode is not axis_mode._InputType.COMPOSITE
                )
            elif mode is axis_mode._InputType.RADIOS_X2:
                self._radioGB.setEnabled(True)
                self._sinogramGB.setEnabled(False)
                self._radioGB.setChecked(True)
            else:
                raise ValueError(f"Nothing implemented for {mode.value}")

    def getCompositeOptions(self) -> dict:
        return {
            "theta": self.getTheta(),
            "oversampling": self.getOversampling(),
            "subsampling_y": self.getSubsamplingY(),
            "take_log": self.getTakeLog(),
        }

    def setCompositeOptions(self, opts: dict) -> None:
        if not isinstance(opts, dict):
            raise TypeError("opts should be an instance of dict")
        for key in opts.keys():
            if key not in ("theta", "oversampling", "subsampling_y", "take_log"):
                raise KeyError(f"{key} is not recogized")
            theta = opts.get("theta", None)
            if theta is not None:
                self.setTheta(theta=theta)
            oversampling = opts.get("oversampling", None)
            if oversampling is not None:
                self.setOversampling(oversampling)
            subsampling_y = opts.get("subsampling_y", None)
            if subsampling_y is not None:
                self.setSubsamplingY(subsampling_y)
            take_log = opts.get("take_log", None)
            if take_log is not None:
                self.setTakeLog(take_log)

    def getTheta(self) -> int:
        return self._thetaSB.value()

    def setTheta(self, theta: int) -> None:
        self._thetaSB.setValue(theta)

    def getOversampling(self) -> int:
        return self._oversamplingSB.value()

    def setOversampling(self, oversampling: int) -> None:
        self._oversamplingSB.setValue(oversampling)

    def getSubsamplingY(self) -> int:
        return self._subsamplingYSB.value()

    def setSubsamplingY(self, subsampling: int) -> None:
        self._subsamplingYSB.setValue(subsampling)

    def getTakeLog(self) -> bool:
        return self._takeLogCB.isChecked()

    def setTakeLog(self, log: bool) -> None:
        self._takeLogCB.setChecked(log)


class _AngleSelectionWidget(qt.QWidget):
    """Group box to select the angle to used for cor calculation
    (0-180, 90-270 or manual)"""

    sigChanged = qt.Signal()
    """signal emitted when the selected angle changed"""

    def __init__(self, parent=None, axis_params=None):
        assert isinstance(axis_params, QAxisRP)
        qt.QWidget.__init__(
            self,
            parent=parent,
        )
        self.setLayout(qt.QVBoxLayout())
        self._groupBoxMode = qt.QGroupBox(
            self, title="Angles to use for axis calculation"
        )
        self._groupBoxMode.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._groupBoxMode)

        self._corButtonsGps = qt.QButtonGroup(parent=self)
        self._corButtonsGps.setExclusive(True)
        self._qrbCOR_0_180 = qt.QRadioButton("0-180", parent=self)
        self._groupBoxMode.layout().addWidget(self._qrbCOR_0_180)
        self._qrbCOR_90_270 = qt.QRadioButton("90-270", parent=self)
        self._qrbCOR_90_270.setToolTip(
            "pick radio closest to angles 90° and "
            "270°. If disable mean that the scan "
            "range is 180°"
        )
        self._groupBoxMode.layout().addWidget(self._qrbCOR_90_270)
        self._qrbCOR_manual = qt.QRadioButton("other", parent=self)
        self._qrbCOR_manual.setVisible(True)
        self._groupBoxMode.layout().addWidget(self._qrbCOR_manual)
        # add all button to the button group
        for b in (self._qrbCOR_0_180, self._qrbCOR_90_270, self._qrbCOR_manual):
            self._corButtonsGps.addButton(b)

        self.setAxisParams(axis_params)

        self._manualFrameSelection = _ManualFramesSelection(self)
        self.layout().addWidget(self._manualFrameSelection)
        self._manualFrameSelection.setVisible(False)

        # connect signal / Slot
        self._corButtonsGps.buttonClicked.connect(self._angleModeChanged)
        self._manualFrameSelection.sigChanged.connect(self._changed)

    def setScan(self, scan: TomwerScanBase):
        if scan is not None:
            self.setScanRange(scan.scan_range)
        self._manualFrameSelection.setScan(scan=scan)

    def setScanRange(self, scanRange):
        if scanRange == 180:
            self._qrbCOR_90_270.setEnabled(False)
            if self._qrbCOR_90_270.isChecked():
                self._qrbCOR_0_180.setChecked(True)

    def setAngleMode(self, mode):
        """

        :param mode: mode to use (can be manual , 90-270 or 0-180)
        """
        assert isinstance(mode, CorAngleMode)
        if mode == CorAngleMode.use_0_180:
            self._qrbCOR_0_180.setChecked(True)
        elif mode == CorAngleMode.use_90_270:
            self._qrbCOR_90_270.setChecked(True)
        else:
            self._qrbCOR_manual.setChecked(True)

    def getAngleMode(self):
        """

        :return: the angle to use for the axis calculation
        :rtype: CorAngleMode
        """
        if self._qrbCOR_90_270.isChecked():
            return CorAngleMode.use_90_270
        elif self._qrbCOR_0_180.isChecked():
            return CorAngleMode.use_0_180
        else:
            return CorAngleMode.manual_selection

    def setAxisParams(self, axis_params):
        old = self.blockSignals(True)
        self._axis_params = axis_params
        # set up
        self.setAngleMode(axis_params.angle_mode)
        self.blockSignals(old)

    def _angleModeChanged(self, *args, **kwargs):
        self._axis_params.angle_mode = self.getAngleMode()
        if self.getAngleMode() is CorAngleMode.manual_selection:
            self._axis_params.angle_mode_extra = (
                self._manualFrameSelection.getFramesUrl()
            )
        else:
            self._axis_params.angle_mode_extra = None
        self._manualFrameSelection.setVisible(
            self.getAngleMode() is CorAngleMode.manual_selection
        )
        self._changed()

    def _changed(self):
        self.sigChanged.emit()


class _ManualFramesSelection(qt.QWidget):
    """Allows to select frame - angle to be used."""

    sigChanged = qt.Signal()
    """Signal emit when the frame selection changes"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())
        self.layout().addWidget(qt.QLabel("frame 1", self), 0, 0, 1, 1)
        self._frame1CB = qt.QComboBox(self)
        self.layout().addWidget(self._frame1CB, 0, 1, 1, 1)

        self.layout().addWidget(qt.QLabel("frame 2", self), 1, 0, 1, 1)
        self._frame2CB = qt.QComboBox(self)
        self.layout().addWidget(self._frame2CB, 1, 1, 1, 1)
        self._flipLRCB = qt.QCheckBox("flip L-R")
        self._flipLRCB.setChecked(True)
        self.layout().addWidget(self._flipLRCB, 1, 2, 1, 1)

        self._flipLRCB.toggled.connect(self._changed)
        self._frame1CB.currentIndexChanged.connect(self._changed)
        self._frame2CB.currentIndexChanged.connect(self._changed)

    def _changed(self):
        self.sigChanged.emit()

    def setScan(self, scan: Union[TomwerScanBase, None]) -> None:
        self._frame1CB.clear()
        self._frame2CB.clear()
        if scan is None:
            return
        current_angle1 = self._getFrame1Angle()
        current_angle2 = self._getFrame2Angle()
        for proj_angle, proj_url in scan.get_proj_angle_url().items():
            try:
                angle = "{:0.3f}".format(float(proj_angle))
            except Exception:
                angle = proj_angle
            self._frame1CB.addItem(angle, proj_url)
            self._frame2CB.addItem(angle, proj_url)

        idx = self._frame1CB.findText(current_angle1)
        if idx >= 0:
            self._frame1CB.setCurrentIndex(idx)

        idx = self._frame2CB.findText(current_angle2)
        if idx >= 0:
            self._frame2CB.setCurrentIndex(idx)

    def getFramesUrl(self, as_txt=False) -> tuple:
        """
        Return a tuple of (frame 1 url, frame 2 url). Url can be None
        """
        if as_txt:
            return self.getFrame1Url().path(), self.getFrame2Url().path()
        else:
            return self.getFrame1Url(), self.getFrame2Url()

    def getFrame1Url(self) -> Optional[DataUrl]:
        return self._frame1CB.currentData()

    def getFrame2Url(self) -> Optional[DataUrl]:
        return self._frame2CB.currentData()

    def _getFrame1Angle(self) -> Optional[str]:
        return self._frame1CB.currentText()

    def _getFrame2Angle(self) -> Optional[str]:
        return self._frame2CB.currentText()

    def isFrame2LRFLip(self):
        return self._flipLRCB.isChecked()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import shutil
import tempfile
import unittest
from tomwer.core.utils.scanutils import MockEDF
from ..reconstruction.lamino import LaminoReconstruction
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams


class TestLaminoIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.recons_params = ReconsParams()

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        for input_type in (dict, TomwerScanBase):
            for return_dict in (True, False):
                with self.subTest(
                    return_dict=return_dict,
                    input_type=input_type,
                ):
                    input_obj = self.scan
                    if input_obj is dict:
                        input_obj = input_obj.to_dict()
                    lamino_process = LaminoReconstruction(
                        inputs={
                            "dry_run": True,
                            "return_dict": return_dict,
                            "data": input_obj,
                        }
                    )
                    lamino_process.run()
                    out = lamino_process.outputs.data
                    if return_dict:
                        self.assertTrue(isinstance(out, dict))
                    else:
                        self.assertTrue(isinstance(out, TomwerScanBase))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "12/12/2018"


from tomwer.core.process.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
import time
import logging

_logger = logging.getLogger(__name__)


class Timer(Task, input_names=("data",), output_names=("data",)):
    """
    Simple timer / time out - function
    """

    def __init__(self, varinfo=None, inputs=None, name=None):
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=name)
        if inputs is None:
            inputs = {}
        self.waiting_time = inputs.get("wait", 1)

    @property
    def waiting_time(self):
        return self._waiting_time

    @waiting_time.setter
    def waiting_time(self, wait):
        self._waiting_time = wait

    def run(self):
        scan = self.inputs.data
        if type(scan) is dict:
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            scan = scan
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                "scan is expected to be a dict or an instance "
                "of TomwerScanBase. Not {}".format(type(scan))
            )
        time.sleep(self.waiting_time)
        if self._return_dict:
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""some utils relative to PyHST
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/01/2017"

import os
import re
import fnmatch


def _findPyHSTVersions(directory):
    """Try to get the PyHST executables"""

    pyhst_exe = []
    pattern = re.compile("PyHST2_[0-9][0-9][0-9][0-9][a-zA]")
    for f in os.listdir(directory):
        if pattern.match(f):
            pyhst_exe.append(f)

    return pyhst_exe


def _getPyHSTDir():
    """
    :return: the directory where the PyHST executable are"""
    if "PYHST_DIR" in os.environ:
        return os.environ["PYHST_DIR"]
    elif os.path.isdir("/usr/bin/"):
        return "/usr/bin/"
    else:
        return None


def _getPyHSTFromEnvVariable():
    pyhst_exe = []
    pattern = "PyHST2_*"
    for f in os.environ.keys():
        if fnmatch.fnmatch(f, pattern):
            pyhst_exe.append(f)

    return pyhst_exe



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse
from silx.gui import qt
import signal
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.gui.visualization.dataviewer import DataViewer
from tomwer.core.utils.resource import increase_max_number_file
import logging


logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="For EDF acquisition: provide folder path, for HDF5 / nexus "
        "provide the master file",
        default=None,
    )
    parser.add_argument(
        "entry",
        help="For Nexus files: entry in the master file",
        default=None,
        nargs="?",
    )

    options = parser.parse_args(argv[1:])

    scan = ScanFactory.create_scan_object(
        scan_path=options.scan_path, entry=options.entry
    )
    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    window = DataViewer(parent=None)
    window.setDisplayMode("radios")
    window.setWindowTitle("tomwer: scan-viewer")
    window.setWindowIcon(icons.getQIcon("tomwer"))
    window.setScan(scan)
    splash.finish(window)
    window.show()

    qt.QApplication.restoreOverrideCursor()
    app.aboutToQuit.connect(window.close)
    exit(app.exec_())


def getinputinfo():
    return "tomwer scan-viewer [file_path] [[file_entry]]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define a set of folders to be emitted to the next box.
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "05/07/2017"


from tomwer.core.signal import Signal
from tomwer.core.process.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
import logging

logger = logging.getLogger(__name__)


class _ScanListPlaceHolder(
    Task, optional_input_names=("data",), output_names=("data",)
):
    """For now data can only be a single element and not a list.
    This must be looked at.
    Also when part of an ewoks graph 'data' is mandatory which is not the class
    when part of a orange workflow. Those can be added interactively"""

    def run(self):
        self.outputs.data = self.inputs.data


class _ScanList(Task, optional_input_names=("data",), output_names=("data",)):
    """Process to list a set of scan"""

    scanReady = Signal(TomwerScanBase)

    def __init__(self, varinfo=None, inputs=None, label=None):
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=label)
        self._scanIDs = {}
        """associate path (key) to TomoBase object"""

    def run(self):
        """function to launch if is the first box to be executed"""
        self._sendList()
        res = []
        for scanID, scan in self._scanIDs.items():
            if self._return_dict:
                res.append(scan.to_dict())
            else:
                res.append(scan)
        self.outputs.data = res

    def set_configuration(self, properties):
        if "_scanIDs" in properties:
            self.setScanIDs(properties["_scanIDs"])
        else:
            raise ValueError("scansID no included in the widget properties")

    def setScanIDs(self, list_of_scan):
        for folder in list_of_scan:
            self.add(folder)

    def add(self, folder):
        """Add a folder to the list

        :param folder: the path to the folder for the scan to add
        :type folder: Union[str, :class:`.TomoBase`]
        """
        if isinstance(folder, TomwerScanBase):
            _scan_obj = folder
        else:
            try:
                _scan_obj = ScanFactory.create_scan_object(folder)
            except ValueError as e:
                logger.warning(e)
                return
        self._scanIDs[str(_scan_obj)] = _scan_obj
        return _scan_obj

    def remove(self, folder):
        """Remove a folder to the list

        :param str folder: the path to the folder for the scan to add
        """
        if folder in self._scanIDs:
            del self._scanIDs[folder]

    def clear(self):
        """clear the list"""
        self._scanIDs.clear()

    def _sendList(self):
        for scanID, scan in self._scanIDs.items():
            self.scanReady.emit(scan)

    def length(self):
        return len(self._scanIDs)


class ScanList(_ScanList):
    """For now to avoid multiple inheritance from QObject with the process widgets
    we have to define two classes. One only for the QObject inheritance
    """

    data = _ScanList.scanReady



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"


import os
import shutil
import tempfile
import unittest

from tomwer.core.process.reconstruction.axis import AxisProcess
from tomwer.core.process.reconstruction.axis.params import AxisRP
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.core.process.reconstruction.ftseries import Ftseries
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from tomwer.core.process.reconstruction.darkref.params import DKRFRP, Method
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.utils.scanutils import MockEDF


class TestTomoReconsWorkflow(unittest.TestCase):
    """Test several tomography reconstruction workflow and make sure they
    are correctly processed. Focus on ReconsParams values"""

    def setUp(self):
        super().setUp()
        # axis reconstruction parameters
        self._axis_rp = AxisRP()
        self._axis_rp_id = id(self._axis_rp)

        # darkref reconstruction parameters
        self._darkref_rp = DKRFRP()
        self._darkref_rp_id = id(self._darkref_rp)

        # ftseries reconstruction parameters
        self._ftseries_rp = ReconsParams()
        self._ftseries_rp_id = id(self._ftseries_rp)

        self._check_default_values()

        # change the value of calculation mode and axis mode
        self._axis_rp.mode = AxisMode.centered
        self._darkref_rp.dark_calc_method = Method.median
        self._ftseries_rp.dkrf.dark_calc_method = Method.median

        # define scans
        self._scans_dir = tempfile.mkdtemp()

        def create_scan(dataset):
            _folder = os.path.join(self._scans_dir, dataset)
            os.makedirs(_folder)
            MockEDF.fastMockAcquisition(_folder, n_radio=10)
            return EDFTomoScan(scan=_folder)

        self.scan_a = create_scan("scanA")
        self.scan_b = create_scan("scanB")
        self.scan_01_dft_rp = ReconsParams()
        self.scan_10_dft_rp = ReconsParams()
        self._recons_params_dft = {
            self.scan_a.path: self.scan_01_dft_rp,
            self.scan_b.path: self.scan_10_dft_rp,
        }
        self.scans = (self.scan_a, self.scan_b)

    def tearDown(self):
        shutil.rmtree(self._scans_dir)
        super().tearDown()

    def _check_default_values(self):
        assert self._darkref_rp.dark_calc_method is Method.average
        assert self._ftseries_rp.dkrf.dark_calc_method is Method.average
        assert id(self._darkref_rp) != id(self._ftseries_rp)

    def _process_axis(self, scan, initial_recons_params):
        """execute the axis process and check results"""
        self._axis_process = AxisProcess(
            inputs={
                "axis_params": self._axis_rp,
                "data": scan,
            }
        )
        self._axis_process.run()
        if initial_recons_params is not None:
            # make sure a new instance of tomo_recons_param has not been changed
            self.assertTrue(
                id(scan.ftseries_recons_params) is id(self._recons_params_dft[scan])
            )

        # test value of the parameter has been modify
        self.assertTrue(scan.axis_params.mode is AxisMode.centered)

    def _process_darkref(self, scan, initial_recons_params):
        """execute the darkref process and check results"""
        self._darkref_process = DarkRefs(
            inputs={
                "force_sync": True,
                "data": scan,
                "dark_ref_params": self._darkref_rp,
            }
        )

        self._darkref_process.run()
        if initial_recons_params is not None:
            # make sure a new instance of tomo_recons_param has not been changed
            self.assertTrue(
                id(scan.ftseries_recons_params) is id(self._recons_params_dft[scan])
            )

        # test that modifications only affect the axis reconstruction parameters
        if initial_recons_params is None:
            self.assertTrue(scan.ftseries_recons_params.paganin is None)

        self.assertTrue(
            scan.ftseries_recons_params.dkrf.dark_calc_method is Method.median
        )

    def _process_ftseries(self, scan, initial_recons_params):
        """execute the ftseries process and check results"""
        self._ftseries_process = Ftseries(
            inputs={
                "force_sync": True,
                "mock_mode": True,
                "data": scan,
                "pyhst2_params": self._ftseries_rp,
            }
        )
        self.assertTrue(
            id(self._ftseries_process.recons_params) == self._ftseries_rp_id
        )

        self.assertEqual(id(self._ftseries_rp), self._ftseries_rp_id)
        self.assertEqual(id(self._ftseries_process.recons_params), self._ftseries_rp_id)

        self._ftseries_process.run()
        if initial_recons_params is not None:
            # make sure a new instance of tomo_recons_param has not been changed
            self.assertTrue(
                id(scan.ftseries_recons_params) is id(self._recons_params_dft[scan])
            )

        if initial_recons_params is None:
            self.assertTrue(scan.ftseries_recons_params.paganin is not None)

        # ftseries should take values from the dkrf and axis objects
        self.assertTrue(
            id(self._ftseries_process.recons_params) == self._ftseries_rp_id
        )
        # make sure ftseries loaded values from axis and dkrf
        self.assertTrue(
            self._ftseries_rp.dkrf.dark_calc_method
            == scan.ftseries_recons_params.dkrf.dark_calc_method
        )
        # make sure dkrf value is unchanged
        self.assertTrue(
            scan.ftseries_recons_params.dkrf.dark_calc_method is Method.median
        )

    def test_axis_darkrefs_ftseries(self):
        """Test that the workflow axis -> darkrefs -> ftseries is valid
        regarding the reconstruction parameters used"""
        for initial_recons_params in (None,):
            with self.subTest(initial_recons_params=initial_recons_params):
                if initial_recons_params is not None:
                    for (key, value) in initial_recons_params.items():
                        key.ftseries_recons_params = value
                        assert id(value) is id(key.ftseries_recons_params)

                for scan in self.scans:
                    assert isinstance(scan, EDFTomoScan)
                    self._process_axis(
                        scan=scan, initial_recons_params=initial_recons_params
                    )
                    self._process_darkref(
                        scan=scan, initial_recons_params=initial_recons_params
                    )
                    self._process_ftseries(
                        scan=scan, initial_recons_params=initial_recons_params
                    )
                    scan.ftseries_recons_params.to_unique_recons_set()

    def test_axis_ftseries(self):
        """Test that the workflow axis -> ftseries sis valid regarding
        the reconstruction parameters used"""
        for initial_recons_params in (None,):
            with self.subTest(initial_recons_params=initial_recons_params):
                if initial_recons_params is not None:
                    for (key, value) in initial_recons_params.items():
                        key.ftseries_recons_params = value
                        assert id(value) is id(key.ftseries_recons_params)

                for scan in self.scans:
                    assert isinstance(scan, EDFTomoScan)
                    self._process_axis(
                        scan=scan, initial_recons_params=initial_recons_params
                    )
                    self._process_ftseries(
                        scan=scan, initial_recons_params=initial_recons_params
                    )
                    scan.ftseries_recons_params.to_unique_recons_set()

    def test_darkrefs_ftseries(self):
        """Test that the workflow darkrefs -> ftseries sis valid regarding
        the reconstruction parameters used"""
        for initial_recons_params in (None,):
            with self.subTest(initial_recons_params=initial_recons_params):
                if initial_recons_params is not None:
                    for (key, value) in initial_recons_params.items():
                        key.ftseries_recons_params = value
                        assert id(value) is id(key.ftseries_recons_params)

                for scan in self.scans:
                    assert isinstance(scan, EDFTomoScan)
                    self._process_darkref(
                        scan=scan, initial_recons_params=initial_recons_params
                    )
                    scan.ftseries_recons_params.to_unique_recons_set()
                    self._process_ftseries(
                        scan=scan, initial_recons_params=initial_recons_params
                    )
                    scan.ftseries_recons_params.to_unique_recons_set()



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "28/08/2020"


from silx.gui import qt
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.reconstruction.nabu import nabuslices, nabuvolume
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils import isLowOnMemory
from tomwer.io.utils import format_stderr_stdout
from ..processingstack import FIFO, ProcessingThread
from processview.core.superviseprocess import SuperviseProcess
from processview.core.manager import ProcessManager, DatasetState
import gc
import logging

_logger = logging.getLogger(__name__)


class NabuSliceProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisProcess` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, parent=None, process_id=None):
        qt.QObject.__init__(self, parent=parent)
        FIFO.__init__(self, process_id=process_id)
        self._dry_run = False

    def _process(self, scan, configuration, callback=None):
        _logger.info("Nabu slice stack is processing {}".format(str(scan)))
        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=DatasetState.ON_GOING,
        )

        self._scan_currently_computed = scan
        assert isinstance(scan, TomwerScanBase)
        self._computationThread.finished.connect(self._end_threaded_computation)

        if isOnLbsram(scan) and isLowOnMemory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = "low memory, skip nabu reconstruction for", scan.path
            _logger.processSkipped(mess)
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=DatasetState.SKIPPED,
            )
            self._end_threaded_computation()
        else:
            self._computationThread.init(scan=scan, configuration=configuration)
            self._computationThread.setDryRun(self._dry_run)
            # need to manage connect before starting it because
            self._computationThread.start()

    def _end_threaded_computation(self, callback=None):
        self._computationThread.finished.disconnect(self._end_threaded_computation)
        super()._end_threaded_computation(callback=callback)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _SliceProcessingThread(process_id=process_id)

    def setDryRun(self, dry_run):
        self._dry_run = dry_run


class NabuVolumeProcessStack(NabuSliceProcessStack):
    """Implementation of the `.AxisProcess` but having a stack for treating
    scans and making computation in threads"""

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _VolumeProcessingThread(process_id=process_id)


class _SliceProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of nabu reconstruction
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self._scan = None
        self._future_scan = None
        self._configuration = None
        self._dry_run = False

    @property
    def future_scan(self):
        return self._future_scan

    def setDryRun(self, dry_run):
        self._dry_run = dry_run

    def init(self, scan, configuration):
        self._scan = scan
        self._configuration = configuration

    def run(self):
        self.sigComputationStarted.emit()
        mess = "Start nabu slice(s) reconstruction of {}".format(str(self._scan))
        _logger.processStarted(mess)
        ProcessManager().notify_dataset_state(
            dataset=self._scan,
            process=self,
            state=DatasetState.ON_GOING,
            details=mess,
        )

        try:
            (
                succeed,
                stdouts,
                stderrs,
                configs,
                self._future_scan,
            ) = nabuslices.run_slices_reconstruction(
                config=self._configuration,
                scan=self._scan,
                dry_run=self._dry_run,
                process_id=self.process_id,
            )
        except Exception as e:
            mess = "Fail to compute slices for {}. Reason is {}.".format(
                str(self._scan), e
            )
            _logger.processFailed(mess)
            ProcessManager().notify_dataset_state(
                dataset=self._scan,
                process=self,
                state=DatasetState.FAILED,
                details=mess,
            )
            self._future_scan = None
        else:
            index = self._scan.pop_process_index()
            if isinstance(self._scan, EDFTomoScan):
                entry = None
            else:
                entry = (
                    self._scan.entry
                )  # hotfix / FIXME: otherwise can fail to open the file to
                # register the process. Might be fix since using HDF5File
            gc.collect()
            try:
                nabuslices.NabuSlices._register_process(
                    process_file=self._scan.process_file,
                    entry=entry,
                    configuration=self._configuration,
                    results={},
                    process=nabuslices.NabuSlices,
                    process_index=index,
                    overwrite=True,
                )
            except Exception as e:
                _logger.warning(f"Fail to register NabuSlices process. Reason is {e}")

            # TODO: check output files with the tomoscan validator ?
            if not succeed:
                mess = "Slices computation for {} failed.".format(str(self._scan))
                state = DatasetState.FAILED
                _logger.processFailed(mess)
            else:
                state = DatasetState.SUCCEED
                mess = "Slices computed for {}.".format(str(self._scan))
                _logger.processSucceed(mess)
            elmts = [
                format_stderr_stdout(stderr=stderr, stdout=stdout)
                for stderr, stdout, config in zip(stderrs, stdouts, configs)
            ]
            elmts.insert(0, mess)
            details = "\n".join(elmts)

            ProcessManager().notify_dataset_state(
                dataset=self._scan,
                process=self,
                state=state,
                details=details,
            )


class _VolumeProcessingThread(_SliceProcessingThread):
    """
    Thread use to execute the processing of nabu reconstruction
    """

    def run(self):
        self.sigComputationStarted.emit()
        mess = "Start nabu volume reconstruction of {}".format(str(self._scan))
        _logger.processStarted(mess)
        ProcessManager().notify_dataset_state(
            dataset=self._scan,
            process=self,
            state=DatasetState.ON_GOING,
            details=mess,
        )
        try:
            (
                succeed,
                stdouts,
                stderrs,
                configs,
                self._future_scan,
            ) = nabuvolume.run_volume_reconstruction(
                config=self._configuration,
                scan=self._scan,
                dry_run=self._dry_run,
                process_id=self.process_id,
            )
        except Exception as e:
            self._future_scan = None
            mess = "Fail to compute volume for {}. Reason is {}".format(
                str(self._scan), e
            )
            _logger.processFailed(mess)
            ProcessManager().notify_dataset_state(
                dataset=self._scan,
                process=self,
                state=DatasetState.FAILED,
                details=mess,
            )
        else:
            index = self._scan.pop_process_index()
            if isinstance(self._scan, EDFTomoScan):
                entry = None
            else:
                entry = self._scan.entry
            gc.collect()
            try:
                nabuvolume.NabuVolume._register_process(
                    process_file=self._scan.process_file,
                    entry=entry,
                    configuration=self._configuration,
                    results={},
                    process=nabuvolume.NabuVolume,
                    process_index=index,
                    overwrite=True,
                )
            except Exception as e:
                _logger.warning(f"Fail to register NabuVolume process. Reason is {e}")
            elmts = [
                format_stderr_stdout(stderr=stderr, stdout=stdout, config=config)
                for stderr, stdout, config in zip(stderrs, stdouts, configs)
            ]

            if not succeed:
                mess = "Volume computed for {} failed.".format(str(self._scan))
                _logger.processFailed(mess)
                state = DatasetState.FAILED
            else:
                mess = "Volume computed for {}.".format(str(self._scan))
                _logger.processSucceed(mess)
                state = DatasetState.SUCCEED

            elmts.insert(0, mess)
            details = "\n".join(elmts)

            ProcessManager().notify_dataset_state(
                dataset=self._scan,
                process=self,
                state=state,
                details=details,
            )
